import logging
import pytest
import time
from copy import copy

from ophyd import (
    PVPositioner,
    PVPositionerPC,
    EpicsSignal,
    EpicsSignalRO,
    Component as Cpt,
    get_cl,
    Kind,
    PVPositionerIsClose,
    PVPositionerDone,
)
from ophyd.utils.epics_pvs import _wait_for_value

logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger("ophyd.pv_positioner").setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)


def tearDownModule():
    logger.debug("Cleaning up")
    logging.getLogger("ophyd.pv_positioner").setLevel(logging.INFO)
    logger.setLevel(logging.INFO)


def test_not_subclassed():
    # can't instantiate it on its own
    with pytest.raises(TypeError):
        PVPositioner("prefix")

    with pytest.raises(TypeError):
        PVPositionerPC("prefix")


def test_no_setpoint_or_readback():
    class MyPositioner(PVPositioner):
        pass

    with pytest.raises(ValueError):
        MyPositioner()


def test_setpoint_but_no_done():
    class MyPositioner(PVPositioner):
        setpoint = Cpt(EpicsSignal, ".VAL")

    with pytest.raises(ValueError):
        MyPositioner()


@pytest.mark.motorsim
def test_pvpos(motor):
    class MyPositioner(PVPositioner):
        """Setpoint, readback, done, stop. No put completion"""

        setpoint = Cpt(EpicsSignal, ".VAL")
        readback = Cpt(EpicsSignalRO, ".RBV")
        done = Cpt(EpicsSignalRO, ".MOVN")
        stop_signal = Cpt(EpicsSignal, ".STOP")

        stop_value = 1
        done_value = 0

    m = MyPositioner(motor.prefix, name="pos_no_put_compl")
    m.wait_for_connection()

    m.read()

    motor.move(0.1, wait=True)
    time.sleep(1)
    assert m.position == 0.1

    m.stop()
    m.limits

    repr(m)
    str(m)

    mc = copy(m)
    assert mc.describe() == m.describe()

    m.read()


@pytest.mark.motorsim
def test_put_complete_setpoint_only(motor):
    logger.info("--> PV Positioner, using put completion and a DONE pv")

    class MyPositioner(PVPositionerPC):
        """Setpoint only"""

        setpoint = Cpt(EpicsSignal, ".VAL")

    pos = MyPositioner(motor.prefix, name="pc_setpoint_done")
    print(pos.describe())
    pos.wait_for_connection()

    pos.read()
    high_lim = pos.setpoint.high_limit
    try:
        pos.check_value(high_lim + 1)
    except ValueError as ex:
        logger.info("Check value for single failed, as expected (%s)", ex)
    else:
        raise ValueError("check_value should have failed")

    stat = pos.move(1, wait=False)
    logger.info("--> post-move request, moving=%s", pos.moving)

    while not stat.done:
        logger.info("--> moving... %s error=%s", stat, stat.error)
        time.sleep(0.1)

    pos.move(-1, wait=True)
    assert not pos.moving


@pytest.mark.motorsim
def test_put_complete_setpoint_readback_done(motor):
    class MyPositioner(PVPositionerPC):
        """Setpoint, readback, done, stop. Put completion"""

        setpoint = Cpt(EpicsSignal, ".VAL")
        readback = Cpt(EpicsSignalRO, ".RBV")
        done = Cpt(EpicsSignalRO, ".MOVN")
        done_value = 0

    pos = MyPositioner(
        motor.prefix, name="pos_no_put_compl", settle_time=0.1, timeout=25.0
    )
    print(pos.describe())
    pos.wait_for_connection()

    assert pos.settle_time == 0.1
    assert pos.timeout == 25.0
    pos.read()
    high_lim = pos.setpoint.high_limit
    try:
        pos.check_value(high_lim + 1)
    except ValueError as ex:
        logger.info("Check value for single failed, as expected (%s)", ex)
    else:
        raise ValueError("check_value should have failed")

    stat = pos.move(1, wait=False)
    assert stat.timeout == pos.timeout
    logger.info("--> post-move request, moving=%s", pos.moving)

    while not stat.done:
        logger.info("--> moving... %s error=%s", stat, stat.error)
        time.sleep(0.1)

    pos.move(-1, wait=True)
    assert not pos.moving


@pytest.mark.motorsim
def test_put_complete_setpoint_readback(motor):
    class MyPositioner(PVPositionerPC):
        """Setpoint, readback, put completion. No done pv."""

        setpoint = Cpt(EpicsSignal, ".VAL")
        readback = Cpt(EpicsSignalRO, ".RBV")

    pos = MyPositioner(motor.prefix, name="pos_put_compl")
    print(pos.describe())
    pos.wait_for_connection()

    stat = pos.move(2, wait=False)
    logger.info("--> post-move request, moving=%s", pos.moving)

    while not stat.done:
        logger.info("--> moving... %s", stat)
        time.sleep(0.1)

    pos.move(0, wait=True)
    logger.info("--> synchronous move request, moving=%s", pos.moving)

    time.sleep(0.1)
    print("read", pos.read())
    assert not pos.moving


def test_pvpositioner_with_fake_motor(fake_motor_ioc):
    def callback(sub_type=None, timestamp=None, value=None, **kwargs):
        logger.info("[callback] [%s] (type=%s) value=%s", timestamp, sub_type, value)

    def done_moving(value=0.0, **kwargs):
        logger.info("Done moving %s", kwargs)

    cl = get_cl()
    # ensure we start at 0 for this simple test
    cl.caput(fake_motor_ioc.pvs["setpoint"], 0)
    cl.caput(fake_motor_ioc.pvs["actuate"], 1)
    time.sleep(0.5)

    class MyPositioner(PVPositioner):
        """Setpoint, readback, no put completion. No done pv."""

        setpoint = Cpt(EpicsSignal, fake_motor_ioc.pvs["setpoint"])
        readback = Cpt(EpicsSignalRO, fake_motor_ioc.pvs["readback"])
        actuate = Cpt(EpicsSignal, fake_motor_ioc.pvs["actuate"])
        stop_signal = Cpt(EpicsSignal, fake_motor_ioc.pvs["stop"])
        done = Cpt(EpicsSignal, fake_motor_ioc.pvs["moving"])

        actuate_value = 1
        stop_value = 1
        done_value = 0

    pos = MyPositioner("", name="pv_pos_fake_mtr")
    print("fake mtr", pos.describe())
    pos.wait_for_connection()

    pos.subscribe(callback, event_type=pos.SUB_DONE)
    pos.subscribe(callback, event_type=pos.SUB_READBACK)

    logger.info("---- test #1 ----")
    logger.info("--> move to 1")
    pos.move(1, timeout=5)
    assert pos.position == 1
    logger.info("--> move to 0")
    pos.move(0, timeout=5)
    assert pos.position == 0

    logger.info("---- test #2 ----")
    logger.info("--> move to 1")
    pos.move(1, wait=False)
    time.sleep(0.5)
    logger.info("--> stop")
    pos.stop()
    logger.info("--> sleep")
    time.sleep(1)
    logger.info("--> move to 0")
    pos.move(0, wait=False, moved_cb=done_moving)
    logger.info("--> post-move request, moving=%s", pos.moving)
    time.sleep(2)

    pos.read()
    repr(pos)
    str(pos)


def test_hints(fake_motor_ioc):
    class MyPositioner(PVPositioner):
        """Setpoint, readback, no put completion. No done pv."""

        setpoint = Cpt(EpicsSignal, fake_motor_ioc.pvs["setpoint"])
        readback = Cpt(EpicsSignalRO, fake_motor_ioc.pvs["readback"])
        actuate = Cpt(EpicsSignal, fake_motor_ioc.pvs["actuate"])
        stop_signal = Cpt(EpicsSignal, fake_motor_ioc.pvs["stop"])
        done = Cpt(EpicsSignal, fake_motor_ioc.pvs["moving"])

        actuate_value = 1
        stop_value = 1
        done_value = 1

    motor = MyPositioner("", name="pv_pos_fake_mtr")

    desc = motor.describe()
    f_hints = motor.hints["fields"]
    assert len(f_hints) > 0
    for k in f_hints:
        assert k in desc

    motor.readback.kind = Kind.hinted
    assert motor.hints == {"fields": ["pv_pos_fake_mtr_readback"]}

    assert motor.hints["fields"] == f_hints


def test_pv_positioner_is_close(signal_test_ioc):
    class MyPositioner(PVPositionerIsClose):
        setpoint = Cpt(EpicsSignal, signal_test_ioc.pvs["read_write"])
        readback = Cpt(EpicsSignal, signal_test_ioc.pvs["pair_set"])

        atol = 0.1

    motor = MyPositioner("", name="pv_pos_is_close_fake_motor")
    setpoint_status = motor.setpoint.set(0)
    readback_status = motor.readback.set(0)
    setpoint_status.wait(timeout=1)
    readback_status.wait(timeout=1)
    goal = motor.position + 10
    status = motor.set(goal)
    _wait_for_value(motor.setpoint, goal, atol=0.01, timeout=1)
    _wait_for_value(motor.done, 0, timeout=1)
    with pytest.raises(TimeoutError):
        _wait_for_value(motor.done, 1, timeout=1)
    motor.readback.put(goal / 2)
    with pytest.raises(TimeoutError):
        _wait_for_value(motor.done, 1, timeout=1)
    motor.readback.put(goal + motor.atol / 2)
    status.wait(timeout=1)
    assert status.done
    assert status.success
    assert motor.done.get() == 1


def test_pv_positioner_done(signal_test_ioc):
    # Catch done going to 0 and back to 1
    motor = PVPositionerDone(
        signal_test_ioc.pvs["read_write"], name="pv_pos_done_fake_motor"
    )
    motor.setpoint.set(0).wait(timeout=1)
    done_values = []

    def accumulate_done_values(value, **kwargs):
        done_values.append(value)

    motor.done.subscribe(accumulate_done_values, run=False)
    motor.set(motor.position + 10).wait()
    assert done_values == [0, 1]



import inspect
import os
import sys
import re
from collections import namedtuple

from ..utils.epics_pvs import records_from_db
from .detectors import AreaDetector
from .base import EpicsSignalWithRBV


StubInfo = namedtuple("StubInfo", ("signal_type", "record"))


def get_prop_name(pv):
    """Get a property name from the camel-case AreaDetector PV name"""
    # If it's all capital letters and underscores, then just convert
    # to lower-case and that's it
    m = re.match("^[A-Z0-9_]+$", pv)
    if m:
        return pv.lower()

    # If the name starts with a bunch of capital letters, use
    # all but the last one as one word
    # e.g., TESTOne -> test_one
    m = re.match("^([A-Z0-9]+)", pv)
    if m:
        start_caps = m.groups()[0]
        pv = pv[len(start_caps) :]
        pv = "".join([start_caps[:-1].lower(), "_", start_caps[-1].lower(), pv])

    # Get all groups of caps or lower-case
    # e.g., AAAbbbCCC -> 'AAA', 'bbb, 'CCC'
    split = re.findall("([a-z0-9:]+|[:A-Z0-9]+)", pv)
    ret = []

    # Put them all back together, removing single-letter splits
    for s in split:
        if ret and len(ret[-1]) == 1:
            ret[-1] += s
        else:
            ret.append(s)

    return "_".join(ret).lower()


def _suffixes_from_device(devcls):
    """Get all suffixes from a device, given its class"""
    for cpt, attr in devcls._sig_attrs.items():
        if hasattr(cpt, "defn"):
            items = [(cls, suffix) for cls, suffix, kwargs in cpt.defn.values()]
        elif hasattr(cpt, "suffix"):
            items = [(cpt.cls, cpt.suffix)]
        else:
            items = []

        for cls, suffix in items:
            yield suffix
            if issubclass(cls, EpicsSignalWithRBV):
                yield "{}_RBV".format(suffix)


def get_stub_info(db_file, macros=None, base_class=AreaDetector):
    """Stub out a new AreaDetector directly from a database file

    Yields lines of code from the generated class definition.

    Parameters
    ----------
    db_file : str
        The database filename to load from
    macros : list, optional
        List of macros to remove, defaults to just $(P)$(R)
    base_class : class
        The base class - required to determine which signals _not_ to add.
        Defaults to AreaDetector
    """
    if not inspect.isclass(base_class):
        raise ValueError(
            "base_class should be a class, got {!r} instead" "".format(base_class)
        )

    if macros is None:
        macros = ("$(P)$(R)",)

    records = [record for rtype, record in records_from_db(db_file)]

    def remove_macros(record):
        for macro in macros:
            record = record.replace(macro, "")
        return record

    records = [remove_macros(record) for record in records]

    # Get all the signals on the base class and remove those from
    # the list.

    base_recs = _suffixes_from_device(base_class)
    records = set(records) - set(base_recs)
    rbv_records = [record for record in records if record.endswith("_RBV")]
    records -= set(rbv_records)
    rbv_only = [
        record for record in rbv_records if record.rsplit("_", 1)[0] not in records
    ]

    records = set(records).union(rbv_only)
    records = list(records)

    for record in sorted(records):
        if record in rbv_only:
            type_ = "ro"
        else:
            rbv_pv = "%s_RBV" % record
            has_rbv = rbv_pv in rbv_records
            if has_rbv:
                type_ = "with_rbv"
            else:
                type_ = "rw"

        yield StubInfo(record=record, signal_type=type_)


def create_detector_stub(
    db_file,
    macros=None,
    base_class=AreaDetector,
    property_name_fcn=None,
    det_name=None,
    cpt_class="C",
    signal_rbv_class="SignalWithRBV",
    signal_rw_class="EpicsSignal",
    signal_ro_class="EpicsSignalRO",
):
    """Stub out a new AreaDetector camera directly from a database file

    Yields lines of code from the generated class definition.

    Parameters
    ----------
    db_file : str
        The database filename to load from
    macros : list, optional
        List of macros to remove, defaults to just $(P)$(R)
    base_class : class
        The base class - required to determine which signals _not_ to add.
        Defaults to AreaDetector
    property_name_fcn : callable, optional
        Function to create property name from a pv. Signature:
            # def name_func(pv):
            #     return 'prop_name'
    det_name : str, optional
        The detector name which will be the class name
        Defaults to DatabaseFilenameDetector
    cpt_class : str
        Component class name to use. Default 'C'
    signal_rbv_class: str
        Class to use when a signal has a separate setpoint/readback.
        Default 'SignalWithRBV'
    signal_rw_class: str
        Class to use when a signal has only a single setpoint/readback PV.
        Default 'EpicsSignal'
    signal_ro_class: str
        Default 'EpicsSignalRO'
        Class to use when only a readback PV exists.
    """

    if det_name is None:
        det_name = os.path.split(db_file)[1]
        det_name = os.path.splitext(det_name)[0]
        det_name = "%sDetectorCam" % det_name

    yield "class {}({}):".format(det_name, base_class.__name__)

    if property_name_fcn is None:
        property_name_fcn = get_prop_name

    yield "    _html_docs = ['']"

    stub_info = get_stub_info(db_file, macros=macros, base_class=base_class)

    class_map = {
        "ro": signal_ro_class,
        "rw": signal_rw_class,
        "with_rbv": signal_rbv_class,
    }

    for info in sorted(stub_info):
        prop_name = property_name_fcn(info.record)
        cls = class_map[info.signal_type]
        info = dict(
            prop_name=prop_name, cpt_class=cpt_class, cls=cls, record=info.record
        )
        yield ("    {prop_name} = {cpt_class}({cls}, {record!r})" "".format(**info))


def stub_templates(path, **kwargs):
    """Stub out a new AreaDetector directly from all database files in a path"""
    for fn in os.listdir(path):
        full_fn = os.path.join(path, fn)
        if fn.endswith(".db") or fn.endswith(".template"):
            yield from create_detector_stub(full_fn, **kwargs)


if __name__ == "__main__":
    stub_templates("/epics/support/areaDetector/1-9-1/ADApp/Db/")
    sys.exit(0)



import time
import logging

import pytest
import operator
import numpy as np


from copy import copy

from ophyd.epics_motor import EpicsMotor
from ophyd.pseudopos import PseudoPositioner, PseudoSingle

from ophyd import Component as C
from ophyd.utils import ExceptionBundle

logger = logging.getLogger(__name__)


def setUpModule():
    logging.getLogger("ophyd.pseudopos").setLevel(logging.DEBUG)


def tearDownModule():

    logger.debug("Cleaning up")
    logging.getLogger("ophyd.pseudopos").setLevel(logging.INFO)


motor_recs = [
    "XF:31IDA-OP{Tbl-Ax:X1}Mtr",
    "XF:31IDA-OP{Tbl-Ax:X2}Mtr",
    "XF:31IDA-OP{Tbl-Ax:X3}Mtr",
    "XF:31IDA-OP{Tbl-Ax:X4}Mtr",
    "XF:31IDA-OP{Tbl-Ax:X5}Mtr",
    "XF:31IDA-OP{Tbl-Ax:X6}Mtr",
]


class Pseudo3x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, "", limits=(-10, 10), egu="a")
    pseudo2 = C(PseudoSingle, "", limits=(-10, 10), egu="b")
    pseudo3 = C(PseudoSingle, "", limits=None, egu="c")
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(
            real1=-pseudo_pos.pseudo1,
            real2=-pseudo_pos.pseudo2,
            real3=-pseudo_pos.pseudo3,
        )

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(
            pseudo1=real_pos.real1, pseudo2=real_pos.real2, pseudo3=real_pos.real3
        )


class Pseudo1x3(PseudoPositioner):
    pseudo1 = C(PseudoSingle, limits=(-10, 10))
    real1 = C(EpicsMotor, motor_recs[0])
    real2 = C(EpicsMotor, motor_recs[1])
    real3 = C(EpicsMotor, motor_recs[2])

    def forward(self, pseudo_pos):
        pseudo_pos = self.PseudoPosition(*pseudo_pos)
        # logger.debug('forward %s', pseudo_pos)
        return self.RealPosition(
            real1=-pseudo_pos.pseudo1,
            real2=-pseudo_pos.pseudo1,
            real3=-pseudo_pos.pseudo1,
        )

    def inverse(self, real_pos):
        real_pos = self.RealPosition(*real_pos)
        # logger.debug('inverse %s', real_pos)
        return self.PseudoPosition(pseudo1=-real_pos.real1)


class FaultyStopperEpicsMotor(EpicsMotor):
    def stop(self, *, success=False):
        raise RuntimeError("Expected exception")


class FaultyPseudo1x3(Pseudo1x3):
    real1 = C(FaultyStopperEpicsMotor, motor_recs[0])


def test_onlypseudo():
    # can't instantiate it on its own
    with pytest.raises(TypeError):
        PseudoPositioner("prefix")


def test_position_wrapper():
    pseudo = Pseudo3x3("", name="mypseudo", concurrent=False)

    test_pos = pseudo.PseudoPosition(pseudo1=1, pseudo2=2, pseudo3=3)
    extra_kw = dict(a=3, b=4, c=6)

    # positional arguments
    assert pseudo.to_pseudo_tuple(1, 2, 3, **extra_kw) == (test_pos, extra_kw)
    # sequence
    assert pseudo.to_pseudo_tuple((1, 2, 3), **extra_kw) == (test_pos, extra_kw)
    # correct type
    assert pseudo.to_pseudo_tuple(test_pos, **extra_kw) == (test_pos, extra_kw)
    # kwargs
    assert pseudo.to_pseudo_tuple(pseudo1=1, pseudo2=2, pseudo3=3, **extra_kw) == (
        test_pos,
        extra_kw,
    )

    # too many positional arguments
    with pytest.raises(ValueError):
        pseudo.to_pseudo_tuple(1, 2, 3, 4)
    # valid kwargs, but passing in args too
    with pytest.raises(ValueError):
        pseudo.to_pseudo_tuple(1, pseudo1=1, pseudo2=2, pseudo3=3)


@pytest.mark.motorsim
def test_multi_sequential():
    pseudo = Pseudo3x3("", name="mypseudo", concurrent=False)
    pseudo.wait_for_connection()

    assert pseudo.egu == "a, b, c"

    pos2 = pseudo.PseudoPosition(pseudo1=0, pseudo2=0, pseudo3=0)
    pseudo.set(pos2, wait=True)
    time.sleep(1.0)
    pos1 = pseudo.PseudoPosition(pseudo1=0.1, pseudo2=0.2, pseudo3=0.3)
    pseudo.set(pos1, wait=True)

    pseudo.real1.set(0, wait=True)
    pseudo.real2.set(0, wait=True)
    pseudo.real3.set(0, wait=True)

    pseudo.pseudo1.stop()

    pseudo.real3.set(0, wait=True)


@pytest.mark.motorsim
def test_faulty_stopper():
    pseudo = FaultyPseudo1x3("", name="mypseudo", concurrent=False)
    pseudo.wait_for_connection()

    with pytest.raises(ExceptionBundle):
        # smoke-testing for coverage
        pseudo.pseudo1.stop()


def test_limits():
    pseudo = Pseudo3x3("", name="mypseudo", concurrent=True)
    assert pseudo.limits == ((-10, 10), (-10, 10), (0, 0))
    assert pseudo.low_limit == (-10, -10, 0)
    assert pseudo.high_limit == (10, 10, 0)


@pytest.mark.motorsim
def test_read_describe():
    pseudo = Pseudo3x3("", name="mypseudo", concurrent=True)
    pseudo.wait_for_connection()
    desc_dict = pseudo.describe()
    desc_keys = [
        "source",
        "upper_ctrl_limit",
        "lower_ctrl_limit",
        "shape",
        "dtype",
        "units",
    ]

    for key in desc_keys:
        assert key in desc_dict["mypseudo_pseudo3"]

    read_dict = pseudo.read()
    read_keys = ["value", "timestamp"]
    for key in read_keys:
        assert key in read_dict["mypseudo_pseudo3"]

    assert pseudo.read().keys() == pseudo.describe().keys()


@pytest.mark.motorsim
def test_multi_concurrent():
    def done(status, **kwargs):
        logger.debug("** Finished moving (%s, %s)", status, kwargs)

    pseudo = Pseudo3x3(
        "", name="mypseudo", concurrent=True, settle_time=0.1, timeout=25.0
    )
    assert pseudo.sequential is False
    assert pseudo.concurrent is True
    assert pseudo.settle_time == 0.1
    assert pseudo.timeout == 25.0
    pseudo.wait_for_connection()
    assert pseudo.connected
    assert tuple(pseudo.pseudo_positioners) == (
        pseudo.pseudo1,
        pseudo.pseudo2,
        pseudo.pseudo3,
    )
    assert tuple(pseudo.real_positioners) == (pseudo.real1, pseudo.real2, pseudo.real3)

    logger.info("Move to (.2, .2, .2), which is (-.2, -.2, -.2) for real " "motors")
    pseudo.set(pseudo.PseudoPosition(0.2, 0.2, 0.2), wait=True)
    logger.info("Position is: %s (moving=%s)", pseudo.position, pseudo.moving)
    pseudo.check_value((2, 2, 2))
    pseudo.check_value(pseudo.PseudoPosition(2, 2, 2))
    try:
        pseudo.check_value((2, 2, 2, 3))
    except ValueError as ex:
        logger.info("Check value failed, as expected (%s)", ex)
    real1 = pseudo.real1
    pseudo1 = pseudo.pseudo1
    try:
        pseudo.check_value((real1.high_limit + 1, 2, 2))
    except ValueError as ex:
        logger.info("Check value failed, as expected (%s)", ex)

    ret = pseudo.set((2, 2, 2), wait=False, moved_cb=done)
    assert ret.settle_time == 0.1
    count = 0
    while not ret.done:
        logger.info("Pos=%s %s (err=%s)", pseudo.position, ret, ret.error)
        count += 1
        if count > 1000:
            raise Exception
        time.sleep(0.1)
    logger.info("Single pseudo axis: %s", pseudo1)

    pseudo1.set(0, wait=True, timeout=5)
    assert pseudo1.target == 0
    pseudo1.sync()
    assert pseudo1.target == pseudo1.position
    # coverage
    pseudo1._started_moving

    try:
        pseudo1.check_value(real1.high_limit + 1)
    except ValueError as ex:
        logger.info("Check value for single failed, as expected (%s)", ex)

    logger.info("Move pseudo1 to 0, position=%s", pseudo.position)
    logger.info("pseudo1 = %s", pseudo1.position)

    def single_sub(**kwargs):
        # logger.info('Single sub: %s', kwargs)
        pass

    pseudo1.subscribe(single_sub, pseudo1.SUB_READBACK)
    ret = pseudo1.set(1, wait=False)
    assert pseudo.timeout == ret.timeout
    count = 0
    while not ret.done:
        logger.info(
            "pseudo1.pos=%s Pos=%s %s (err=%s)",
            pseudo1.position,
            pseudo.position,
            ret,
            ret.error,
        )
        count += 1
        if count > 20:
            raise Exception
        time.sleep(0.1)

    logger.info(
        "pseudo1.pos=%s Pos=%s %s (err=%s)",
        pseudo1.position,
        pseudo.position,
        ret,
        ret.error,
    )
    copy(pseudo)
    pseudo.read()
    pseudo.describe()
    pseudo.read_configuration()
    pseudo.describe_configuration()
    repr(pseudo)
    str(pseudo)
    pseudo.pseudo1.read()
    pseudo.pseudo1.describe()
    pseudo.pseudo1.read_configuration()
    pseudo.pseudo1.describe_configuration()


@pytest.mark.motorsim
def test_single_pseudo():
    logger.info("------- Sequential, single pseudo positioner")
    pos = Pseudo1x3("", name="mypseudo", concurrent=False)
    pos.wait_for_connection()
    reals = pos._real

    logger.info("Move to .2, which is (-.2, -.2, -.2) for real motors")
    pos.set((0.2,), wait=True)
    logger.info("Position is: %s (moving=%s)", pos.position, pos.moving)
    logger.info("Real positions: %s", [real.position for real in reals])

    logger.info("Move to -.2, which is (.2, .2, .2) for real motors")
    pos.set((-0.2,), wait=True)
    logger.info("Position is: %s (moving=%s)", pos.position, pos.moving)
    logger.info("Real positions: %s", [real.position for real in reals])

    copy(pos)
    pos.read()
    pos.describe()
    repr(pos)
    str(pos)


@pytest.mark.parametrize(
    "inpargs,inpkwargs,expected_position,expected_kwargs",
    [
        ((1, 2, 3), {}, (1, 2, 3), {}),
        ((1, 2), {}, (1, 2, -3), {}),
        ((1,), {}, (1, -2, -3), {}),
        (((1, 2, 3),), {}, (1, 2, 3), {}),
        (([1, 2],), {}, (1, 2, -3), {}),
        (((1,),), {}, (1, -2, -3), {}),
        ((), {"pseudo1": 1, "pseudo2": 2, "pseudo3": 3}, (1, 2, 3), {}),
        ((), {"pseudo1": 1, "pseudo2": 2}, (1, 2, -3), {}),
        ((), {"pseudo1": 1}, (1, -2, -3), {}),
        ((), {"pseudo1": 1, "wait": True}, (1, -2, -3), {"wait": True}),
        (({"pseudo1": 1, "pseudo2": 2, "pseudo3": 3},), {}, (1, 2, 3), {}),
        (({"pseudo1": 1, "pseudo2": 2},), {}, (1, 2, -3), {}),
        (({"pseudo1": 1},), {}, (1, -2, -3), {}),
        (
            ({"pseudo1": 1, "wait": True},),
            {"timeout": None},
            (1, -2, -3),
            {"wait": True, "timeout": None},
        ),
        ((1, 2, 3), {"timeout": 1}, (1, 2, 3), {"timeout": 1}),
        (((1, 2, 3),), {"timeout": 1}, (1, 2, 3), {"timeout": 1}),
    ],
)
def test_pseudo_position_input_3x3(
    hw, inpargs, inpkwargs, expected_position, expected_kwargs
):
    pseudo3x3 = hw.pseudo3x3
    pseudo3x3.real1.set(1)
    pseudo3x3.real2.set(2)
    pseudo3x3.real3.set(3)

    out, extra_kwargs = pseudo3x3.to_pseudo_tuple(*inpargs, **inpkwargs)
    assert out == pseudo3x3.PseudoPosition(*expected_position)
    assert extra_kwargs == expected_kwargs

    pseudo3x3.set(*inpargs, **inpkwargs)
    assert pseudo3x3.position == pseudo3x3.PseudoPosition(*expected_position)


@pytest.mark.parametrize(
    "inpargs,inpkwargs",
    [
        ((1, 2, 3, 5), {}),
        ((1, 2, 3), {"pseudo1": 1}),
        ((1, 2, 3), {"pseudo2": 1}),
        ((1,), {"pseudo2": 1, "pseudo3": 1}),
        ((1, 2), {"pseudo3": 1}),
    ],
)
def test_pseudo_position_fail_3x3(hw, inpargs, inpkwargs):
    pseudo3x3 = hw.pseudo3x3
    with pytest.raises(ValueError):
        pseudo3x3.to_pseudo_tuple(*inpargs, **inpkwargs)


@pytest.mark.parametrize(
    "inpargs,inpkwargs,expected_position,expected_kwargs",
    [
        ((1, 2, 3), {}, (1, 2, 3), {}),
        ((1, 2), {}, (1, 2, 3), {}),
        ((1,), {}, (1, 2, 3), {}),
        (((1, 2, 3),), {}, (1, 2, 3), {}),
        (([1, 2],), {}, (1, 2, 3), {}),
        (((1,),), {}, (1, 2, 3), {}),
        ((), {"real1": 1, "real2": 2, "real3": 3}, (1, 2, 3), {}),
        ((), {"real1": 1, "real2": 2}, (1, 2, 3), {}),
        ((), {"real1": 1}, (1, 2, 3), {}),
        ((), {"real1": 1, "foo": "bar"}, (1, 2, 3), {"foo": "bar"}),
        (({"real1": 1, "real2": 2, "real3": 3},), {}, (1, 2, 3), {}),
        (({"real1": 1, "real2": 2},), {}, (1, 2, 3), {}),
        (({"real1": 1},), {}, (1, 2, 3), {}),
        (
            ({"real1": 1, "foo": "bar"},),
            {"baz": "buz"},
            (1, 2, 3),
            {"foo": "bar", "baz": "buz"},
        ),
        ((1, 2, 3), {"foo": "bar"}, (1, 2, 3), {"foo": "bar"}),
    ],
)
def test_real_position_input_3x3(
    hw, inpargs, inpkwargs, expected_position, expected_kwargs
):
    pseudo3x3 = hw.pseudo3x3
    pseudo3x3.real1.set(1)
    pseudo3x3.real2.set(2)
    pseudo3x3.real3.set(3)

    out, extra_kwargs = pseudo3x3.to_real_tuple(*inpargs, **inpkwargs)
    assert out == pseudo3x3.RealPosition(*expected_position)
    assert extra_kwargs == expected_kwargs


@pytest.mark.parametrize(
    "inpargs,inpkwargs",
    [
        ((1, 2, 3, 5), {}),
        ((1, 2, 3), {"real1": 1}),
        ((1, 2, 3), {"real2": 1}),
        ((1,), {"real2": 1, "real3": 1}),
        ((1, 2), {"real3": 1}),
        (({"real3": 1, "foo": "bar"},), {"foo": "bizz"}),
        ((), {}),
    ],
)
def test_real_position_fail_3x3(hw, inpargs, inpkwargs):
    pseudo3x3 = hw.pseudo3x3
    with pytest.raises(ValueError):
        pseudo3x3.to_real_tuple(*inpargs, **inpkwargs)


def test_single_pseudo_with_sim(hw):
    logger.info("------- Sequential, single pseudo positioner")
    pos = hw.pseudo1x3

    reals = pos._real

    logger.info("Move to .2, which is (-.2, -.2, -.2) for real motors")
    pos.set((0.2,), wait=True)
    logger.info("Position is: %s (moving=%s)", pos.position, pos.moving)
    logger.info("Real positions: %s", [real.position for real in reals])

    logger.info("Move to -.2, which is (.2, .2, .2) for real motors")
    pos.set((-0.2,), wait=True)
    logger.info("Position is: %s (moving=%s)", pos.position, pos.moving)
    logger.info("Real positions: %s", [real.position for real in reals])

    copy(pos)
    pos.read()
    pos.describe()
    repr(pos)
    str(pos)


@pytest.mark.parametrize("typ", ("to_real_tuple", "to_pseudo_tuple"))
@pytest.mark.parametrize("op", (operator.sub, operator.add))
@pytest.mark.parametrize(
    "a,b",
    [((0, 0, 0), (1, 1, 1)), ((1, 0, 1), (1, 1, 1)), ((9, 0, 0.3), (0.3, 0.1, 0.5))],
)
def test_pseudo_math(hw, a, b, op, typ):
    pos = hw.pseudo3x3
    a, _ = getattr(pos, typ)(a)
    b, _ = getattr(pos, typ)(b)

    # TODO switch to np asserts
    expected = op(np.asarray(a), np.asarray(b))
    assert (np.asarray(op(a, b)) == expected).all()
    assert (np.asarray(op(a, tuple(b))) == expected).all()
    assert (np.asarray(op(a, list(b))) == expected).all()
    assert (np.asarray(op(a, b._asdict())) == expected).all()
    assert (np.asarray(op(a, {})) == a).all()
    assert abs(op(a, b)) == np.sqrt(np.sum(expected**2))


def test_pseudo_hints(hw):
    pos = hw.pseudo3x3

    for j in (1, 2, 3):
        p = getattr(pos, "pseudo{}".format(j))
        assert p.hints["fields"] == [p.readback.name]
        p.readback.name = "aardvark{}".format(j)
        assert p.hints["fields"] == [p.readback.name]

    expected_fields = [
        getattr(pos, "pseudo{}".format(j)).readback.name for j in (1, 2, 3)
    ]
    assert pos.hints["fields"] == expected_fields



import logging

from collections import OrderedDict

from .ophydobj import Kind
from .signal import EpicsSignal, EpicsSignalRO
from .device import Device
from .device import (
    Component as Cpt,
    DynamicDeviceComponent as DDCpt,
    FormattedComponent as FCpt,
)

logger = logging.getLogger(__name__)


def _scaler_fields(cls, attr_base, field_base, range_, **kwargs):
    defn = OrderedDict()
    for i in range_:
        attr = "{attr}{i}".format(attr=attr_base, i=i)
        suffix = "{field}{i}".format(field=field_base, i=i)
        defn[attr] = (cls, suffix, kwargs)

    return defn


class EpicsScaler(Device):
    """SynApps Scaler Record interface"""

    # tigger + trigger mode
    count = Cpt(EpicsSignal, ".CNT", trigger_value=1, kind=Kind.omitted)
    count_mode = Cpt(EpicsSignal, ".CONT", string=True, kind=Kind.config)

    # delay from triggering to starting counting
    delay = Cpt(EpicsSignal, ".DLY", kind=Kind.config)
    auto_count_delay = Cpt(EpicsSignal, ".DLY1", kind=Kind.config)

    # the data
    channels = DDCpt(
        _scaler_fields(EpicsSignalRO, "chan", ".S", range(1, 33), kind=Kind.hinted)
    )
    names = DDCpt(
        _scaler_fields(EpicsSignal, "name", ".NM", range(1, 33), kind=Kind.config)
    )

    time = Cpt(EpicsSignal, ".T", kind=Kind.config)
    freq = Cpt(EpicsSignal, ".FREQ", kind=Kind.config)

    preset_time = Cpt(EpicsSignal, ".TP", kind=Kind.config)
    auto_count_time = Cpt(EpicsSignal, ".TP1", kind=Kind.config)

    presets = DDCpt(
        _scaler_fields(EpicsSignal, "preset", ".PR", range(1, 33), kind=Kind.omitted)
    )
    gates = DDCpt(
        _scaler_fields(EpicsSignal, "gate", ".G", range(1, 33), kind=Kind.omitted)
    )

    update_rate = Cpt(EpicsSignal, ".RATE", kind=Kind.omitted)
    auto_count_update_rate = Cpt(EpicsSignal, ".RAT1", kind=Kind.omitted)

    egu = Cpt(EpicsSignal, ".EGU", kind=Kind.config)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.stage_sigs.update([("count_mode", 0)])


class ScalerChannel(Device):

    # TODO set up monitor on this to automatically change the name
    chname = FCpt(EpicsSignal, "{self.prefix}.NM{self._ch_num}", kind=Kind.config)
    s = FCpt(
        EpicsSignalRO,
        "{self.prefix}.S{self._ch_num}",
        kind=Kind.hinted,
        auto_monitor=False,
    )
    preset = FCpt(EpicsSignal, "{self.prefix}.PR{self._ch_num}", kind=Kind.config)
    gate = FCpt(
        EpicsSignal, "{self.prefix}.G{self._ch_num}", string=True, kind=Kind.config
    )

    def __init__(self, prefix, ch_num, **kwargs):
        self._ch_num = ch_num

        super().__init__(prefix, **kwargs)
        self.match_name()

    def match_name(self):
        self.s.name = self.chname.get()


def _sc_chans(attr_fix, id_range):
    defn = OrderedDict()
    for k in id_range:
        defn["{}{:02d}".format(attr_fix, k)] = (
            ScalerChannel,
            "",
            {"ch_num": k, "kind": Kind.normal},
        )
    return defn


class ScalerCH(Device):

    # The data
    channels = DDCpt(_sc_chans("chan", range(1, 33)))

    # tigger + trigger mode
    count = Cpt(EpicsSignal, ".CNT", trigger_value=1, kind=Kind.omitted)
    count_mode = Cpt(EpicsSignal, ".CONT", string=True, kind=Kind.config)

    # delay from triggering to starting counting
    delay = Cpt(EpicsSignal, ".DLY", kind=Kind.config)
    auto_count_delay = Cpt(EpicsSignal, ".DLY1", kind=Kind.config)

    time = Cpt(EpicsSignal, ".T")
    freq = Cpt(EpicsSignal, ".FREQ", kind=Kind.config)

    preset_time = Cpt(EpicsSignal, ".TP", kind=Kind.config)
    auto_count_time = Cpt(EpicsSignal, ".TP1", kind=Kind.config)

    update_rate = Cpt(EpicsSignal, ".RATE", kind=Kind.omitted)
    auto_count_update_rate = Cpt(EpicsSignal, ".RAT1", kind=Kind.omitted)

    egu = Cpt(EpicsSignal, ".EGU", kind=Kind.config)

    def match_names(self):
        for s in self.channels.component_names:
            getattr(self.channels, s).match_name()

    def select_channels(self, chan_names=None):
        """Select channels based on the EPICS name PV

        Parameters
        ----------
        chan_names : Iterable[str] or None

            The names (as reported by the ``channel.chname`` signal)
            of the channels to select.
            If ``None``, select **all** channels named in the EPICS scaler.
        """
        self.match_names()
        name_map = {}
        for s in self.channels.component_names:
            scaler_channel = getattr(self.channels, s)
            nm = scaler_channel.s.name  # as defined in self.match_names()
            if len(nm) > 0:
                name_map[nm] = s

        if chan_names is None:
            chan_names = name_map.keys()

        read_attrs = []
        for ch in chan_names:
            try:
                read_attrs.append(name_map[ch])
            except KeyError:
                raise RuntimeError(
                    "The channel {} is not configured "
                    "on the scaler.  The named channels are "
                    "{}".format(ch, tuple(name_map))
                )

        self.channels.kind = Kind.normal
        self.channels.read_attrs = list(read_attrs)
        self.channels.configuration_attrs = list(read_attrs)
        for ch in read_attrs[1:]:
            getattr(self.channels, ch).s.kind = Kind.hinted



import logging
import uuid

import pytest

from types import SimpleNamespace

from ophyd import (
    set_cl,
    EpicsMotor,
    Signal,
    EpicsSignal,
    EpicsSignalRO,
    Component as Cpt,
)
from ophyd.utils.epics_pvs import AlarmSeverity, AlarmStatus

logger = logging.getLogger(__name__)


@pytest.fixture()
def hw(tmpdir):
    from ophyd.sim import hw

    return hw(str(tmpdir))


@pytest.fixture(params=["caproto", "pyepics"], autouse=True, scope="session")
def cl_selector(request):
    cl_name = request.param
    if cl_name == "caproto":
        pytest.importorskip("caproto")
        logging.getLogger("caproto.bcast").setLevel("INFO")
    elif cl_name == "pyepics":
        pytest.importorskip("epics")
    set_cl(cl_name)
    yield
    set_cl()


class CustomAlarmEpicsSignalRO(EpicsSignalRO):
    alarm_status = AlarmStatus.NO_ALARM
    alarm_severity = AlarmSeverity.NO_ALARM


class TestEpicsMotor(EpicsMotor):
    user_readback = Cpt(CustomAlarmEpicsSignalRO, ".RBV", kind="hinted")
    high_limit_switch = Cpt(Signal, value=0, kind="omitted")
    low_limit_switch = Cpt(Signal, value=0, kind="omitted")
    direction_of_travel = Cpt(Signal, value=0, kind="omitted")
    high_limit_value = Cpt(EpicsSignal, ".HLM", kind="config")
    low_limit_value = Cpt(EpicsSignal, ".LLM", kind="config")

    @user_readback.sub_value
    def _pos_changed(self, timestamp=None, value=None, **kwargs):
        """Callback from EPICS, indicating a change in position"""
        super()._pos_changed(timestamp=timestamp, value=value, **kwargs)


@pytest.fixture(scope="function")
def motor(request, cleanup):
    sim_pv = "XF:31IDA-OP{Tbl-Ax:X1}Mtr"

    motor = TestEpicsMotor(sim_pv, name="epicsmotor", settle_time=0.1, timeout=10.0)
    cleanup.add(motor)

    print("Created EpicsMotor:", motor)
    motor.wait_for_connection()
    motor.low_limit_value.put(-100, wait=True)
    motor.high_limit_value.put(100, wait=True)
    motor.set(0).wait()

    return motor


@pytest.fixture(scope="module")
def ad_prefix():
    "AreaDetector prefix"
    # prefixes = ['13SIM1:', 'XF:31IDA-BI{Cam:Tbl}']
    prefixes = ["ADSIM:"]

    for prefix in prefixes:
        test_pv = prefix + "TIFF1:PluginType_RBV"
        try:
            sig = EpicsSignalRO(test_pv)
            sig.wait_for_connection(timeout=2)
        except TimeoutError:
            ...
        else:
            print("areaDetector detected with prefix:", prefix)
            return prefix
        finally:
            sig.destroy()
    raise pytest.skip("No areaDetector IOC running")


@pytest.fixture(scope="function")
def prefix():
    "Random PV prefix for a server"
    return str(uuid.uuid4())[:8] + ":"


@pytest.fixture(scope="function")
def fake_motor_ioc(prefix, request):
    name = "Fake motor IOC"
    pvs = dict(
        setpoint=f"{prefix}setpoint",
        readback=f"{prefix}readback",
        moving=f"{prefix}moving",
        actuate=f"{prefix}actuate",
        stop=f"{prefix}stop",
        step_size=f"{prefix}step_size",
    )

    pytest.importorskip("caproto.tests.conftest")
    from caproto.tests.conftest import run_example_ioc

    process = run_example_ioc(
        "ophyd.tests.fake_motor_ioc",
        request=request,
        pv_to_check=pvs["setpoint"],
        args=("--prefix", prefix, "--list-pvs", "-v"),
    )
    return SimpleNamespace(
        process=process, prefix=prefix, name=name, pvs=pvs, type="caproto"
    )


@pytest.fixture(scope="function")
def signal_test_ioc(prefix, request):
    name = "test_signal IOC"
    pvs = dict(
        read_only=f"{prefix}read_only",
        read_write=f"{prefix}read_write",
        pair_set=f"{prefix}pair_set",
        pair_rbv=f"{prefix}pair_rbv",
        waveform=f"{prefix}waveform",
        bool_enum=f"{prefix}bool_enum",
        alarm_status=f"{prefix}alarm_status",
        set_severity=f"{prefix}set_severity",
        path=f"{prefix}path",
    )

    pytest.importorskip("caproto.tests.conftest")
    from caproto.tests.conftest import run_example_ioc

    process = run_example_ioc(
        "ophyd.tests.signal_ioc",
        request=request,
        pv_to_check=pvs["read_only"],
        args=("--prefix", prefix, "--list-pvs", "-v"),
    )
    return SimpleNamespace(
        process=process, prefix=prefix, name=name, pvs=pvs, type="caproto"
    )


@pytest.fixture(scope="function")
def cleanup(request):
    "Destroy all items added to the list during the finalizer"
    items = []

    class Cleaner:
        def add(self, item):
            items.append(item)

    def clean():
        for item in items:
            print("Destroying", item.name)
            item.destroy()
        items.clear()

    request.addfinalizer(clean)
    return Cleaner()



# vi: ts=4 sw=4


import logging
import time
import threading
import functools
import math

from collections import OrderedDict, namedtuple
from collections.abc import Sequence, Mapping

from .utils import DisconnectedError, ExceptionBundle
from .positioner import PositionerBase, SoftPositioner
from .device import Device, Component as Cpt, Kind, required_for_connection
from .signal import AttributeSignal


logger = logging.getLogger(__name__)


class PseudoSingle(Device, SoftPositioner):
    """A single axis of a PseudoPositioner

    This should not be instantiated on its own, but rather used as a Component
    in a PseudoPositioner subclass.

    Parameters
    ----------
    prefix : str, optional
        The PV prefix, for compatibility with the Device hierarchy
    limits : (low_limit, high_limit)
        User-defined limits for this pseudo axis.
    egu : str, optional
        The engineering units (EGU) for the position
    parent : PseudoPositioner instance
        The instance of the parent PseudoPositioner
    name : str, optional
        The name of the positioner
    source : str, optional
        Metadata indicating the source of this positioner's position. Defaults
        to 'computed'
    target_initial_position : bool, optional
        Use initial inverse-calculated position as the default setpoint/target
        for this axis
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.
    """

    readback = Cpt(AttributeSignal, attr="position", kind=Kind.hinted)
    setpoint = Cpt(AttributeSignal, attr="target", kind=Kind.normal)

    def __init__(
        self,
        prefix="",
        *,
        limits=None,
        egu="",
        parent=None,
        name=None,
        source="computed",
        target_initial_position=False,
        **kwargs,
    ):

        super().__init__(
            prefix=prefix,
            name=name,
            parent=parent,
            limits=limits,
            egu=egu,
            source=source,
            **kwargs,
        )
        # the readback name should default to the name of the positioner
        self.readback.name = self.name

        self._target_initial_position = target_initial_position
        self._target = None

        # The index of this PseudoSingle in the parent PseudoPositioner tuple
        # will be set post-instantiation:
        self._idx = None
        self._parent.subscribe(self._sub_proxy_start, event_type=self.SUB_START)
        self._parent.subscribe(self._sub_proxy_done, event_type=self.SUB_DONE)
        self._parent.subscribe(self._sub_proxy_readback, event_type=self.SUB_READBACK)

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("idx", self._idx)

    def _sub_proxy_start(self, obj=None, **kwargs):
        "Pass through parent callbacks for motion started"
        return self._run_subs(obj=self, **kwargs)

    def _sub_proxy_done(self, obj=None, **kwargs):
        "Pass through parent callbacks for motion started"
        return self._run_subs(obj=self, **kwargs)

    @required_for_connection(description="{device.name} readback subscription")
    def _sub_proxy_readback(self, obj=None, value=None, **kwargs):
        """Parent callbacks including a position value will be filtered through
        this function and re-broadcast using only the relevant position to this
        pseudo axis.
        """
        if hasattr(value, "__getitem__"):
            value = value[self._idx]

        return self._run_subs(obj=self, value=value, **kwargs)

    @property
    def target(self):
        """Last commanded target position"""
        if self._target is None:
            return self.position
        else:
            return self._target

    def sync(self):
        """Synchronize target position with current readback position"""
        self._target = None

    def check_value(self, pos):
        self._parent.check_single(self, pos)

    @property
    def moving(self):
        return self._parent.moving

    @property
    def position(self):
        """The current position of the motor in its engineering units

        Returns
        -------
        position
        """
        return self._parent.position[self._idx]

    def stop(self, *, success=False):
        """Stop motion on the PseudoPositioner"""
        return self._parent.stop(success=success)

    @property
    def _started_moving(self):
        """Has motion started since the motion request?

        This is a property on PseudoSingle, which overrides the default
        behavior of Positioner. It reflects the motion status of the
        PseudoPositioner as a whole.
        """
        return self._parent._started_moving

    @_started_moving.setter
    def _started_moving(self, value):
        # Don't allow the base class to specify whether it has started moving
        pass

    def _setup_move(self, position, status):
        """PseudoSingle.move overrides SoftPositioner move implementation, so
        this method is not called.
        """
        pass

    def move(self, pos, **kwargs):
        """Move this pseudo axis to a specific position.

        See `PseudoPositioner.move_single` for more information.

        Parameters
        ----------
        pos : float
            Position to move to
        kwargs : dict
            Passed onto parent.move_single()
        """
        self._target = pos
        return self._parent.move_single(self, pos, **kwargs)

    def describe(self):
        desc = super().describe()
        low_limit, high_limit = self.limits

        for d in (desc[self.readback.name], desc[self.setpoint.name]):
            d["upper_ctrl_limit"] = high_limit
            d["lower_ctrl_limit"] = low_limit
            d["units"] = self.egu

        return desc


def _position_argument_wrapper(type_):
    """Wrapper to convert positional arguments to a PositionTuple"""

    def wrapper(method):
        @functools.wraps(method)
        def wrapped(self, *args, **kwargs):
            m = {"pseudo": self.to_pseudo_tuple, "real": self.to_real_tuple}[type_]
            pos, new_kwargs = m(*args, **kwargs)

            return method(self, pos, **new_kwargs)

        return wrapped

    return wrapper


real_position_argument = _position_argument_wrapper("real")
pseudo_position_argument = _position_argument_wrapper("pseudo")

_to_position_tuple_usage_info = """Positions can be passed in a number of ways.

As positional arguments:
    pseudo.method(px, py, pz, **kwargs)
As a sequence or PseudoPosition/RealPosition:
    pseudo.method((px, py, pz), **kwargs)
As kwargs:
    pseudo.method(px=1, py=2, pz=3, **kwargs)

"""


def _to_position_tuple(cls, *args, _cur, **kwargs):
    """Convert user-specified arguments to a Position namedtuple and kwargs

    TODO update for _cur

    Example::

        Tuple = namedtuple('Tuple', 'px py pz')

    All of the following will return the same thing::

        t, kwargs = to_position_tuple(Tuple, px, py, pz, a=4)
        t, kwargs = to_position_tuple(Tuple, (px, py, pz), a=4)
        t, kwargs = to_position_tuple(Tuple, Tuple(px, py, pz), a=4)
        t, kwargs = to_position_tuple(Tuple, px=1, py=2, pz=3, a=4)

    t will be Tuple(px, py, pz), and kwargs will be {'a': 4}.

    Parameters
    ----------
    cls : namedtuple
        The position class to use. This is likely a RealPosition or a
        PseudoPosition from a PseudoPositioner.
    args :
        User-specified positional arguments
    kwargs : dict
        User-specified keyword arguments

    Returns
    -------
    position_tuple : cls
        The position tuple
    kwargs : dict
        Keyword arguments not related to the position tuple

    Raises
    ------
    TypeError
        On an empty or invalid namedtuple
    ValueError
        On a mismatch of parameters
    """
    try:
        fields = cls._fields
    except AttributeError:
        raise TypeError("Invalid position tuple")

    if not fields:
        raise TypeError("Invalid position tuple")

    if args:
        if isinstance(args[0], (cls, Sequence)):
            (args,) = args

        elif isinstance(args[0], Mapping):
            (arg,) = args
            if any(k in kwargs for k in arg):
                raise ValueError("overlap between dict arg and kwargs")
            kwargs.update(arg)
            args = tuple()

    # too many args, give up
    if len(args) > len(fields):
        raise ValueError("too many args")

    # have _cur, too few args, and no field names in kwargs, fill out
    elif len(args) > 0:
        if any(f in kwargs for f in fields):
            raise ValueError("can not mix args and kwargs for positions")
        if len(args) == len(fields):
            return cls(*args), kwargs
        else:
            _cur = _cur()
            return cls(*args, *_cur[len(args) :]), kwargs

    # No positional arguments, position described in terms of kwargs
    if not kwargs:
        # no positional arguments or kwargs, just show usage information
        raise ValueError(_to_position_tuple_usage_info)

    missing_fields = [field for field in fields if field not in kwargs]

    if missing_fields:
        _cur = _cur()
        kwargs.update({k: getattr(_cur, k) for k in missing_fields})

    # separate position tuple kwargs from other kwargs
    position_kw = {field: kwargs[field] for field in fields}
    other_kw = {key: value for key, value in kwargs.items() if key not in fields}
    position = cls(**position_kw)
    return position, other_kw


class PseudoPositioner(Device, SoftPositioner):
    """A pseudo positioner which can be comprised of multiple positioners

    Parameters
    ----------
    prefix : str
        The PV prefix for all components of the device
    concurrent : bool, optional
        If set, all real motors will be moved concurrently. If not, they will
        be moved in order of how they were defined initially
    read_attrs : sequence of attribute names
        the components to include in a normal reading (i.e., in ``read()``)
    configuration_attrs : sequence of attribute names
        the components to be read less often (i.e., in
        ``read_configuration()``) and to adjust via ``configure()``
    name : str, optional
        The name of the device
    egu : str, optional
        The user-defined engineering units for the whole PseudoPositioner
    auto_target : bool, optional
        Automatically set the target position of PseudoSingle devices when
        moving to a single PseudoPosition
    parent : instance or None
        The instance of the parent device, if applicable
    settle_time : float, optional
        The amount of time to wait after moves to report status completion
    timeout : float, optional
        The default timeout to use for motion requests, in seconds.
    """

    class __add_sub_mixin:
        """Helper mix-in to make RealPosition and PseudoPosition mathable

        .. warning ::

           Do not use this outside of this purpose.

           No really, don't
        """

        __slots__ = ()

        def __add__(self, other):
            if not isinstance(other, type(self)):
                try:
                    for k in self._fields:
                        other.setdefault(k, 0)
                    other = type(self)(**other)
                except (TypeError, AttributeError):
                    try:
                        other = type(self)(*other)
                    except TypeError:
                        return NotImplemented

            return type(self)(*(s + o for s, o in zip(self, other)))

        def __sub__(self, other):
            if not isinstance(other, type(self)):
                try:
                    for k in self._fields:
                        other.setdefault(k, 0)
                    other = type(self)(**other)
                except (TypeError, AttributeError):
                    try:
                        other = type(self)(*other)
                    except TypeError:
                        return NotImplemented

            return type(self)(*(s - o for s, o in zip(self, other)))

        def __abs__(self):
            return math.sqrt(sum(x * x for x in self))

    def __init__(
        self,
        prefix="",
        *,
        concurrent=True,
        read_attrs=None,
        configuration_attrs=None,
        name,
        egu="",
        auto_target=True,
        **kwargs,
    ):

        self._finished_lock = threading.RLock()
        self._concurrent = bool(concurrent)
        self._finish_thread = None
        self._real_waiting = []
        self._move_queue = []
        self.auto_target = auto_target

        if self.__class__ is PseudoPositioner:
            raise TypeError(
                "PseudoPositioner must be subclassed with the "
                "correct signals set in the class definition."
            )

        super().__init__(
            prefix,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            name=name,
            egu=egu,
            **kwargs,
        )

        self._real = [getattr(self, attr) for attr, cpt in self._get_real_positioners()]
        self._pseudo = [
            getattr(self, attr) for attr, cpt in self._get_pseudo_positioners()
        ]

        if not self._pseudo or not self._real:
            raise ValueError("Must have at least 1 positioner and " "pseudo-positioner")

        if not self._egu:
            # Make the PseudoPositioner units based on the PseudoSingle
            # units
            self._egu = self.composite_egu

        self.RealPosition = self._real_position_tuple()
        self.PseudoPosition = self._pseudo_position_tuple()

        self.log.debug("Real positioners: %s", self._real)
        self.log.debug("Pseudo positioners: %s", self._pseudo)

        for idx, pseudo in enumerate(self._pseudo):
            pseudo._idx = idx

        self._real_cur_pos = OrderedDict((real, None) for real in self._real)

        for real in self._real:
            # Subscribe to events from all the real motors and update the
            # internal state of their position
            self._required_for_connection[real] = f"{real.name} readback position"
            real.subscribe(
                self._real_pos_update, event_type=real.SUB_READBACK, run=True
            )

    @property
    def composite_egu(self):
        """The composite engineering units (EGU) from all PseudoSingles"""
        return ", ".join(pseudo.egu for pseudo in self._pseudo if pseudo.egu)

    @property
    def pseudo_positioners(self):
        """Pseudo positioners instances in a namedtuple

        Returns
        -------
        positioner_instances : PseudoPosition
        """
        return self.PseudoPosition(*self._pseudo)

    @property
    def real_positioners(self):
        """Real positioners instances in a namedtuple

        Returns
        -------
        positioner_instances : RealPosition
        """
        return self.RealPosition(*self._real)

    @classmethod
    def _real_position_tuple(cls):
        """A namedtuple for a real motor position

        This is automatically generated at the class-level for all
        non-PseudoSingle-based positioners.
        """
        cname = cls.__name__ + "RealPos"

        return type(
            cname,
            (
                cls.__add_sub_mixin,
                namedtuple(
                    "_" + cname, [name for name, cpt in cls._get_real_positioners()]
                ),
            ),
            {},
        )

    @classmethod
    def _pseudo_position_tuple(cls):
        """A namedtuple for a pseudo motor position

        This is automatically generated at the class-level for all
        PseudoSingle-based positioners.
        """
        cname = cls.__name__ + "PseudoPos"
        return type(
            cname,
            (
                cls.__add_sub_mixin,
                namedtuple(
                    "_" + cname, [name for name, cpt in cls._get_pseudo_positioners()]
                ),
            ),
            {},
        )

    @classmethod
    def _get_pseudo_positioners(cls):
        """Inspect the components and find the pseudo positioners

        All `PseudoSingle` (and subclassed) components will be returned, by
        default.

        The built-in mechanism to override the list of pseudo positioners on a
        PseudoPositioner is to define '_pseudo' on the class-level.  It should
        be a list of attribute names.

        Yields
        ------
        (attr, component)
        """
        if hasattr(cls, "_pseudo"):
            for pseudo in cls._pseudo:
                yield pseudo, getattr(cls, pseudo)
        else:
            for attr, cpt in cls._sig_attrs.items():
                if issubclass(cpt.cls, PseudoSingle):
                    yield attr, cpt

    @classmethod
    def _get_real_positioners(cls):
        """Inspect the components and find the real positioners

        All `Positioner` components which are not `PseudoSingle`s will be
        returned, by default.

        The built-in mechanism to override the list of real positioners on a
        PseudoPositioner is to define '_real' on the class-level.  It should be
        a list of attribute names. This allows you to group real motors
        logically on the device but not have them included in motions or
        calculations.

        Yields
        ------
        (attr, component)
        """
        if hasattr(cls, "_real"):
            for real in cls._real:
                yield real, getattr(cls, real)
        else:
            for attr, cpt in cls._sig_attrs.items():
                is_pseudo = issubclass(cpt.cls, PseudoSingle)
                is_positioner = issubclass(cpt.cls, PositionerBase)
                if is_positioner and not is_pseudo:
                    yield attr, cpt

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("concurrent", self._concurrent)

    def stop(self, success=False):
        del self._move_queue[:]
        exc_list = []

        for attr in self._sub_devices:
            dev = getattr(self, attr)

            if isinstance(dev, PseudoSingle) or not dev.connected:
                continue

            try:
                dev.stop(success=success)
            except ExceptionBundle as ex:
                exc_list.extend(
                    [
                        ("{}.{}".format(attr, sub_attr), ex)
                        for sub_attr, ex in ex.exceptions.items()
                    ]
                )
            except Exception as ex:
                exc_list.append((attr, ex))
                self.log.exception("Device %s (%s) stop failed", attr, dev)

        if exc_list:
            exc_info = "\n".join(
                "{} raised {!r}".format(attr, ex) for attr, ex in exc_list
            )
            raise ExceptionBundle(
                "{} exception(s) were raised during stop: \n"
                "{}".format(len(exc_list), exc_info),
                exceptions=dict(exc_list),
            )

    def check_single(self, pseudo_single, single_pos):
        """Check if a new position for a single pseudo positioner is valid"""
        idx = pseudo_single._idx
        target = list(self.target)
        target[idx] = single_pos
        return self.check_value(self.PseudoPosition(*target))

    def to_pseudo_tuple(self, *args, **kwargs):
        """Convert arguments to a PseudoPosition namedtuple and kwargs"""
        return _to_position_tuple(
            self.PseudoPosition, *args, **kwargs, _cur=lambda: self.target
        )

    def to_real_tuple(self, *args, **kwargs):
        """Convert arguments to a RealPosition namedtuple and kwargs"""
        return _to_position_tuple(
            self.RealPosition, *args, **kwargs, _cur=lambda: self.real_position
        )

    def check_value(self, pseudo_pos):
        """Check if a new position for all pseudo positioners is valid

        First checks limits against those set for individual pseudo axes.
        Second, calculates forward(pseudo_pos) => real_pos and checks it
        against the real positioners.

        NOTE: If you have limits that are coupled together or are somehow more
        complicated than the above procedure, you should redefine this method
        in your subclass.
        """
        try:
            pseudo_pos = self.PseudoPosition(*pseudo_pos)
        except TypeError as ex:
            raise ValueError(
                "Not all required values for a PseudoPosition: {}"
                "({})".format(self.PseudoPosition._fields, ex)
            )

        for pseudo, pos in zip(self._pseudo, pseudo_pos):
            low, high = pseudo.limits
            if (high > low) and not (low <= pos <= high):
                raise ValueError(
                    "Position is outside of pseudo single limits:"
                    " {}, {} < {} < {}".format(pseudo.name, low, pos, high)
                )

        real_pos = self.forward(pseudo_pos)
        for real, pos in zip(self._real, real_pos):
            real.check_value(pos)

    @property
    def limits(self):
        """All PseudoSingle limits as a namedtuple"""
        # NOTE: overrides SoftPositioner implementation
        return self.PseudoPosition(*(pseudo.limits for pseudo in self._pseudo))

    @property
    def low_limit(self):
        """All PseudoSingle low limits as a namedtuple"""
        # NOTE: overrides SoftPositioner implementation
        return self.PseudoPosition(*(pseudo.low_limit for pseudo in self._pseudo))

    @property
    def high_limit(self):
        """All PseudoSingle high limits as a namedtuple"""
        # NOTE: overrides SoftPositioner implementation
        return self.PseudoPosition(*(pseudo.high_limit for pseudo in self._pseudo))

    @property
    def moving(self):
        return any(pos.moving for pos in self._real)

    @property
    def sequential(self):
        """If sequential is set, motors will move in the sequence they were
        defined in (i.e., in series)
        """
        return not self._concurrent

    @property
    def concurrent(self):
        """If concurrent is set, motors will move concurrently (in parallel)"""
        return self._concurrent

    @property
    def _started_moving(self):
        return any(pos._started_moving for pos in self._real)

    @_started_moving.setter
    def _started_moving(self, value):
        # Don't allow the base class to specify whether it has started moving
        pass

    @property
    def position(self):
        """Pseudo motor position namedtuple"""
        return self.inverse(self.real_position)

    @property
    def real_position(self):
        """Real motor position namedtuple"""
        return self.RealPosition(*self._real_cur_pos.values())

    def _update_position(self):
        """Update the internal position based on all of the real positioners"""
        real_cur_pos = self.real_position
        if None in real_cur_pos:
            raise DisconnectedError("Not all positioners connected")

        initial_position = self._position is None
        calc_pseudo_pos = self.inverse(real_cur_pos)
        self._set_position(calc_pseudo_pos)

        if initial_position:
            for positioner, single_pos in zip(self._pseudo, calc_pseudo_pos):
                if positioner._target_initial_position:
                    positioner._target = single_pos

        return calc_pseudo_pos

    def _real_pos_update(self, obj=None, value=None, **kwargs):
        """Callback: A single real positioner has moved"""
        real = obj
        self._real_cur_pos[real] = value
        # Only update the position if all real motors are connected
        try:
            self._update_position()
        except DisconnectedError:
            pass

        # Now that we have a position for this motor, it is no longer blocking
        # the PseudoPositioner from being marked as connected:
        self._required_for_connection.pop(real, None)

    def _done_moving(self, success=True):
        """Call this when motion has completed.  Runs SUB_DONE subscription."""
        del self._real_waiting[:]
        super()._done_moving(success=success)

    def _real_finished(self, status=None, *, obj=None):
        """Callback: A single real positioner has finished moving.

        Used for asynchronous motion, if all have finished moving then fire a
        callback (via `Positioner._done_moving`)
        """
        with self._finished_lock:
            real = obj
            self.log.debug("Real motor %s finished moving", real.name)

            if real in self._real_waiting:
                self._real_waiting.remove(real)

                if not self._real_waiting:
                    self._done_moving()

    def move_single(self, pseudo, position, **kwargs):
        """Move one PseudoSingle axis to a position

        All other positioners will use their current setpoint/target value, if
        available. Failing that, their current readback value will be used (see
        `PseudoSingle.sync` and `PseudoSingle.target`).

        Parameters
        ----------
        pseudo : PseudoSingle
            PseudoSingle positioner to move
        position : float
            Position only for the PseudoSingle
        kwargs : dict
            Passed onto move
        """
        idx = pseudo._idx
        target = list(self.target)
        target[idx] = position
        return self.move(self.PseudoPosition(*target), **kwargs)

    @property
    def target(self):
        """Last commanded target positions"""
        return self.PseudoPosition(*(pos.target for pos in self._pseudo))

    def _sequential_move(self, real_pos, timeout=None, **kwargs):
        """Move all real positioners to a certain position, in series"""
        self._move_queue[:] = zip(self._real, real_pos)
        pending_status = []
        t0 = time.time()

        def move_next(status=None, obj=None):
            # last motion complete message came from 'obj'
            self.log.debug("[%s:sequential] move_next called", self.name)
            with self._finished_lock:
                if pending_status:
                    last_status = pending_status[-1]
                    if not last_status.success:
                        self.log.error("Failing due to last motion")
                        self._done_moving(success=False)
                        return

                try:
                    real, position = self._move_queue.pop(0)
                except IndexError:
                    self._done_moving(success=True)
                    return

                self.log.debug(
                    "[%s:sequential] Moving next motor: %s", self.name, real.name
                )

                elapsed = time.time() - t0
                if timeout is None:
                    sub_timeout = None
                else:
                    sub_timeout = timeout - elapsed

                self.log.debug(
                    "[%s:sequential] Moving %s to %s (timeout=%s)",
                    self.name,
                    real.name,
                    position,
                    sub_timeout,
                )

                if sub_timeout is not None and sub_timeout < 0:
                    self.log.error("Motion timeout")
                    self._done_moving(success=False)
                else:
                    status = real.move(
                        position,
                        wait=False,
                        timeout=sub_timeout,
                        moved_cb=move_next,
                        **kwargs,
                    )
                    pending_status.append(status)
                    self.log.debug(
                        "[%s:sequential] waiting on %s", self.name, real.name
                    )

        self.log.debug("[%s:sequential] started", self.name)
        move_next()

    def _concurrent_move(self, real_pos, **kwargs):
        """Move all real positioners to a certain position, in parallel"""
        self._real_waiting.extend(self._real)

        for real, value in zip(self._real, real_pos):
            self.log.debug("[concurrent] Moving %s to %s", real.name, value)
            real.move(value, wait=False, moved_cb=self._real_finished, **kwargs)

    @pseudo_position_argument
    def move(self, position, wait=True, timeout=None, moved_cb=None):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Pseudo position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        if self.auto_target:
            # in auto-target mode, we update the setpoints of the PseudoSingles
            # on every motion of the PseudoPositioner
            for positioner, single_pos in zip(self._pseudo, position):
                positioner._target = single_pos
        return super().move(position, wait=wait, timeout=timeout, moved_cb=moved_cb)

    move.__doc__ = SoftPositioner.move.__doc__

    def _setup_move(self, position, status):
        """Move requested to position

        This is a customization of SoftPositioner's _setup_move method which
        is what gets called when a motion request happens.

        Parameters
        ----------
        position : PseudoPosition
            Position to move to (already verified by `check_value`)
        status : MoveStatus
            Status object created by PositionerBase.move()
        """
        # Clear all old statuses for not yet completed real motions
        del self._real_waiting[:]

        timeout = status.timeout
        real_pos = self.forward(position)

        with self._finished_lock:
            # ensure we don't get any motion complete messages before motion
            # setup is finished
            if self.sequential:
                self._sequential_move(real_pos, timeout=timeout)
            else:
                self._concurrent_move(real_pos, timeout=timeout)

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        """Calculate a RealPosition from a given PseudoPosition

        Must be defined on the subclass.

        Parameters
        ----------
        pseudo_pos : PseudoPosition
            The pseudo position input

        Returns
        -------
        real_position : RealPosition
            The real position output
        """
        # return self.RealPosition()
        raise NotImplementedError()

    @real_position_argument
    def inverse(self, real_pos):
        """Calculate a PseudoPosition from a given RealPosition

        Must be defined on the subclass.

        Parameters
        ----------
        real_position : RealPosition
            The real position input

        Returns
        -------
        pseudo_pos : PseudoPosition
            The pseudo position output
        """
        # return self.PseudoPosition()
        raise NotImplementedError()

    @pseudo_position_argument
    def set(self, position, **kwargs):
        """Move to a new position asynchronously

        Parameters
        ----------
        position : PseudoPosition
            Position for the all of the pseudo axes

        Returns
        -------
        status : MoveStatus
        """
        return super().set(position, **kwargs)



import pytest
import logging
from ophyd import Device


logger = logging.getLogger(__name__)


def test_specify_version():
    # Define a versioned Device:
    class MyDevice(Device, version=1, version_type="ioc"):
        ...

    info = MyDevice._class_info_
    assert info == {
        "version": 1,
        "versions": {1: MyDevice},
        "version_type": "ioc",
        "version_of": MyDevice,
    }

    # Define a new version of that Device:
    class MyDevice_V2(MyDevice, version=2, version_of=MyDevice):
        ...

    info = MyDevice_V2._class_info_
    assert info == {
        "version": 2,
        "versions": {1: MyDevice, 2: MyDevice_V2},
        "version_type": "ioc",
        "version_of": MyDevice,
    }

    # Ensure that the original Device has also been updated:
    assert MyDevice._class_info_["versions"] == {1: MyDevice, 2: MyDevice_V2}

    # Define a user device that inherits - but does not define a new version
    class UserDevice(MyDevice_V2):
        ...

    assert UserDevice._class_info_ == {
        "versions": {1: MyDevice, 2: MyDevice_V2},
        "version": 2,
        "version_type": "ioc",
        "version_of": MyDevice,
    }


def test_version_requires_subclass():
    class MyDevice(Device, version=1):
        ...

    with pytest.raises(RuntimeError):

        class UnrelatedDevice(Device, version=2, version_of=MyDevice):
            ...



import logging
import pathlib
import time as ttime

from .. import EpicsSignal

logger = logging.getLogger(__name__)
OS_NAME_TO_PATH_CLASS = {
    "nt": pathlib.PureWindowsPath,
    "posix": pathlib.PurePosixPath,
}

OS_SEPARATORS = {"nt": "\\", "posix": "/"}


def path_compare(path_a, path_b, semantics):
    """
    Compare paths, given OS-specific semantics

    Parameters
    ----------
    path_a : str or pathlib.Path
        The first path
    path_b : str or pathlib.Path
        The second path
    semantics : {'nt', 'posix'}
        The OS name for the path

    Returns
    -------
    result : bool
        Whether the paths are equal or not
    """
    try:
        path_class = OS_NAME_TO_PATH_CLASS[semantics]
    except KeyError:
        raise ValueError(f"Unknown path semantics: {semantics}") from None

    return path_class(path_a) == path_class(path_b)


def set_and_wait_path(signal, val, *, path_semantics, poll_time=0.01, timeout=10):
    """
    Set a path signal to a value and wait until it reads back correctly.

    Parameters
    ----------
    signal : EpicsPathSignal (or any object with `get` and `put`)
        The signal itself
    val : object
        value to set signal to
    path_semantics : {'nt', 'posix'}
        The OS name for the path
    poll_time : float, optional
        how soon to check whether the value has been successfully set
    timeout : float, optional
        maximum time to wait for value to be successfully set

    Raises
    ------
    TimeoutError if timeout is exceeded
    """
    # ensure a Path object is converted to a string before setting
    val = str(val)
    # Make sure val has trailing separator before it's set
    if not any(val.endswith(sep) for sep in OS_SEPARATORS.values()):
        val = val + OS_SEPARATORS[path_semantics]
    signal.put(val)
    deadline = ttime.time() + timeout if timeout is not None else None
    current_value = signal.get()

    while not path_compare(current_value, val, semantics=path_semantics):
        logger.debug(
            "Waiting for %s to be set from %r to %r...", signal.name, current_value, val
        )
        ttime.sleep(poll_time)
        if poll_time < 0.1:
            poll_time *= 2  # logarithmic back-off
        current_value = signal.get()
        if deadline is not None and ttime.time() > deadline:
            raise TimeoutError(
                "Attempted to set %r to value %r and timed "
                "out after %r seconds. Current value is %r."
                % (signal, val, timeout, current_value)
            )


class EpicsPathSignal(EpicsSignal):
    def __init__(self, write_pv, *, path_semantics, string=True, **kwargs):
        """
        An areaDetector-compatible EpicsSignal expecting 2 PVs holding a path

        That is, an EpicsPathSignal uses the areaDetector convention of
        'pvname' being the setpoint and 'pvname_RBV' being the read-back path.

        Operating system-specific path semantics are respected when confirming
        that a :meth:`.set()` operation has completed.
        """
        if write_pv.endswith("_RBV"):
            # Strip off _RBV if it was passed in erroneously
            write_pv = write_pv[:-4]

        self.path_semantics = path_semantics
        if string is not True:
            raise ValueError(
                "Specifying an EpicsPathSignal with string=False" " does not make sense"
            )

        super().__init__(
            write_pv=write_pv, read_pv=f"{write_pv}_RBV", string=True, **kwargs
        )

    @property
    def path_semantics(self):
        return self._path_semantics

    @path_semantics.setter
    def path_semantics(self, value):
        if value not in OS_NAME_TO_PATH_CLASS:
            raise ValueError(
                f"Unknown path semantics: {value}. " f"Options: {OS_NAME_TO_PATH_CLASS}"
            )
        self._path_semantics = value

    def _repr_info(self):
        yield from super()._repr_info()
        yield ("path_semantics", self.path_semantics)

    def _set_and_wait(self, value, timeout):
        """
        Overridable hook for subclasses to override :meth:`.set` functionality.

        This will be called in a separate thread (`_set_thread`), but will not
        be called in parallel.

        Parameters
        ----------
        value : any
            The value
        timeout : float, optional
            Maximum time to wait for value to be successfully set, or None
        """
        return set_and_wait_path(
            self, value, timeout=timeout, path_semantics=self.path_semantics
        )



from __future__ import annotations

import collections
import contextlib
import functools
import inspect
import itertools
import logging
import operator
import textwrap
import time as ttime
import typing
import warnings
from collections import OrderedDict, namedtuple
from collections.abc import Iterable, MutableSequence
from enum import Enum
from typing import (
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .ophydobj import Kind, OphydObject
from .signal import Signal
from .status import DeviceStatus, StatusBase
from .utils import (
    ExceptionBundle,
    RedundantStaging,
    doc_annotation_forwarder,
    getattrs,
    underscores_to_camel_case,
)

A, B = TypeVar("A"), TypeVar("B")
ALL_COMPONENTS = object()
# This attrs are defined at instanitation time and must not
# collide with class attributes.
DEVICE_INSTANCE_ATTRS = {
    "name",
    "parent",
    "component_names",
    "_signals",
    "_sig_attrs",
    "_sub_devices",
}
# These attributes are part of the bluesky interface and cannot be
# used as component names.
DEVICE_RESERVED_ATTRS = {
    "read",
    "describe",
    "trigger",
    "configure",
    "read_configuration",
    "describe_configuration",
    "describe_collect",
    "set",
    "stage",
    "unstage",
    "pause",
    "resume",
    "kickoff",
    "complete",
    "collect",
    "position",
    "stop",
    # from OphydObject
    "subscribe",
    "clear_sub",
    "event_types",
    "root",
    # for back-compat
    "signal_names",
}


class OrderedDictType(Dict[A, B]):
    ...


logger = logging.getLogger(__name__)


class Staged(Enum):
    """Three-state switch"""

    yes = "yes"
    no = "no"
    partially = "partially"


ComponentWalk = namedtuple("ComponentWalk", "ancestors dotted_name item")


K = TypeVar("K", bound=OphydObject)


class Component(typing.Generic[K]):
    """A descriptor representing a device component (or signal)

    Unrecognized keyword arguments will be passed directly to the component
    class initializer.

    Parameters
    ----------
    cls : class
        Class of signal to create.  The required signature of `cls.__init__` is
        (if `suffix` is given)::

            def __init__(self, pv_name, parent=None, **kwargs):

        or (if suffix is None) ::

            def __init__(self, parent=None, **kwargs):

        The class may have a `wait_for_connection()` which is called during the
        component instance creation.

    suffix : str, optional
        The PV suffix, which gets appended onto ``parent.prefix`` to generate
        the final PV that the instance component will bind to.
        Also see ``add_prefix``

    lazy : bool, optional
        Lazily instantiate the signal. If ``False``, the signal will be
        instantiated upon component instantiation.  Defaults to
        ``component.lazy_default``.

    trigger_value : any, optional
        Mark as a signal to be set on trigger. The value is sent to the signal
        at trigger time.

    add_prefix : sequence, optional
        Keys in the kwargs to prefix with the Device PV prefix during creation
        of the component instance.
        Defaults to ``('suffix', 'write_pv', )``

    doc : str, optional
        string to attach to component DvcClass.component.__doc__
    """

    #: Default laziness for the component class.
    lazy_default: ClassVar[bool] = False

    #: The attribute name of the component.
    attr: Optional[str]
    #: The class to instantiate when the device is created.
    cls: Type[K]
    #: Keyword arguments for the device creation.
    kwargs: Dict[str, Any]
    #: Lazily create components on access.
    lazy: bool
    #: PV or identifier suffix.
    suffix: Optional[str]
    #: Documentation string.
    doc: Optional[str]
    #: Value to send on ``trigger()``
    trigger_value: Optional[Any]
    #: The data acquisition kind.
    kind: Kind
    #: Names of kwarg keys to prepend the device PV prefix to.
    add_prefix: Tuple[str, ...]
    #: Subscription name -> subscriptions marked by decorator.
    _subscriptions: DefaultDict[str, List[Callable]]

    def __init__(
        self,
        cls: Type[K],
        suffix: Optional[str] = None,
        *,
        lazy: Optional[bool] = None,
        trigger_value: Optional[Any] = None,
        add_prefix: Optional[Sequence[str]] = None,
        doc: Optional[str] = None,
        kind: Union[str, Kind] = Kind.normal,
        **kwargs,
    ):
        self.attr = None  # attr is set later by the device when known
        self.cls = cls
        self.kwargs = kwargs
        self.lazy = lazy if lazy is not None else self.lazy_default
        self.suffix = suffix
        self.doc = doc
        self.trigger_value = trigger_value  # TODO discuss
        self.kind = Kind[kind.lower()] if isinstance(kind, str) else Kind(kind)
        if add_prefix is None:
            add_prefix = ("suffix", "write_pv")
        self.add_prefix = tuple(add_prefix)
        self._subscriptions = collections.defaultdict(list)

    def _get_class_from_annotation(self) -> Optional[Type[K]]:
        """Get a class from the Component[cls] annotation."""
        annotation = getattr(self, "__orig_class__", None)
        if not annotation:
            return None

        args = typing.get_args(annotation)
        if not args or not len(args) == 1:
            return None
        return args[0]

    def __set_name__(self, owner, attr_name: str):
        self.attr = attr_name
        if self.doc is None:
            self.doc = self.make_docstring(owner)

    @property
    def is_device(self):
        "Does this Component contain a Device?"
        return isinstance(self.cls, type) and issubclass(self.cls, Device)

    @property
    def is_signal(self):
        "Does this Component contain a Signal?"
        return isinstance(self.cls, type) and issubclass(self.cls, Signal)

    def maybe_add_prefix(self, instance, kw, suffix):
        """Add prefix to a suffix if kw is in self.add_prefix

        Parameters
        ----------
        instance : Device
            The instance to extract the prefix to maybe append to the
            suffix from.

        kw : str
            The key of associated with the suffix.  If this key is
            self.add_prefix than prepend the prefix to the suffix and
            return, else just return the suffix.

        suffix : str
            The suffix to maybe have something prepended to.

        Returns
        -------
        str
        """
        if kw in self.add_prefix:
            return f"{instance.prefix}{suffix}"
        return suffix

    def create_component(self, instance):
        "Instantiate the object described by this Component for a Device"
        kwargs = self.kwargs.copy()
        kwargs.update(
            name=f"{instance.name}_{self.attr}",
            kind=instance._component_kinds[self.attr],
            attr_name=self.attr,
        )

        for kw, val in list(kwargs.items()):
            kwargs[kw] = self.maybe_add_prefix(instance, kw, val)

        if self.suffix is not None:
            pv_name = self.maybe_add_prefix(instance, "suffix", self.suffix)
            cpt_inst = self.cls(pv_name, parent=instance, **kwargs)
        else:
            cpt_inst = self.cls(parent=instance, **kwargs)

        if self.lazy and hasattr(self.cls, "wait_for_connection"):
            if getattr(instance, "lazy_wait_for_connection", True):
                cpt_inst.wait_for_connection()

        return cpt_inst

    def make_docstring(self, parent_class):
        "Create a docstring for the Component"
        if self.doc is not None:
            return self.doc

        doc = [
            "{} attribute".format(self.__class__.__name__),
            "::",
            "",
        ]

        doc.append(textwrap.indent(repr(self), prefix=" " * 4))
        doc.append("")
        return "\n".join(doc)

    def __repr__(self):
        repr_dict = self.kwargs.copy()
        repr_dict.pop("read_attrs", None)
        repr_dict.pop("configuration_attrs", None)
        repr_dict["kind"] = self.kind.name

        kw_str = ", ".join(f"{k}={v!r}" for k, v in repr_dict.items())
        suffix = repr(self.suffix) if self.suffix else ""

        component_class = self.cls.__name__
        args = ", ".join(s for s in (component_class, suffix, kw_str) if s)

        this_class = self.__class__.__name__
        return f"{this_class}({args})"

    __str__ = __repr__

    @typing.overload
    def __get__(self, instance: None, owner: type) -> Component[K]:
        ...

    @typing.overload
    def __get__(self, instance: Device, owner: type) -> K:
        ...

    def __get__(
        self,
        instance: Optional[Device],
        owner: type,
    ) -> Union[Component, K]:
        if instance is None:
            return self

        try:
            return instance._signals[self.attr]
        except KeyError:
            return instance._instantiate_component(self.attr)

    def __set__(self, instance, owner):
        raise RuntimeError("Do not use setattr with components; use " "cpt.put(value)")

    def subscriptions(self, event_type):
        """(Decorator) Specify subscriptions callbacks in the Device definition

        Parameters
        ----------
        event_type : str or None
            Event type to subscribe to. `ophyd.Signal` supports at least
            {'value', 'meta'}.  An `event_type` of `None` indicates that the
            default event type for the signal is to be used.

        Returns
        -------
        subscriber : callable
            Callable with signature `subscriber(func)`, where `func` is the
            method to call when the subscription of event_type is fired.
        """

        def subscriber(func):
            self._subscriptions[event_type].append(func)
            if not hasattr(func, "_subscriptions"):
                func._subscriptions = []
            func._subscriptions.append((self, event_type))
            return func

        return subscriber

    def sub_default(self, func):
        "Default subscription decorator"
        return self.subscriptions(None)(func)

    def sub_meta(self, func):
        "Metadata subscription decorator"
        return self.subscriptions("meta")(func)

    def sub_value(self, func):
        "Value subscription decorator"
        return self.subscriptions("value")(func)


class FormattedComponent(Component[K]):
    """A Component which takes a dynamic format string

    This differs from Component in that the parent prefix is not automatically
    added onto the Component suffix. Additionally, `str.format()` style strings
    are accepted, allowing access to Device instance attributes:

    >>> from ophyd import (Component as Cpt, FormattedComponent as FCpt)
    >>> class MyDevice(Device):
    ...     # A normal component, where 'suffix' is added to prefix verbatim
    ...     cpt = Cpt(EpicsSignal, 'suffix')
    ...     # A formatted component, where 'self' refers to the Device instance
    ...     ch = FCpt(EpicsSignal, '{self.prefix}{self._ch_name}')
    ...     # A formatted component, where 'self' is assumed
    ...     ch = FCpt(EpicsSignal, '{prefix}{_ch_name}')
    ...
    ...     def __init__(self, prefix, ch_name=None, **kwargs):
    ...         self._ch_name = ch_name
    ...         super().__init__(prefix, **kwargs)

    >>> dev = MyDevice('prefix:', ch_name='some_channel', name='dev')
    >>> print(dev.cpt.pvname)
    prefix:suffix
    >>> print(dev.ch.pvname)
    prefix:some_channel

    For additional documentation, refer to Component.
    """

    def maybe_add_prefix(self, instance, kw, suffix):
        if kw not in self.add_prefix:
            return suffix

        format_dict = dict(instance.__dict__)
        format_dict["self"] = instance
        return suffix.format(**format_dict)


class DynamicDeviceComponent(Component["Device"]):
    """An Device component that dynamically creates an ophyd Device

    Parameters
    ----------
    defn : OrderedDict
        The definition of all attributes to be created, in the form of::

            defn['attribute_name'] = (SignalClass, pv_suffix, keyword_arg_dict)

        This will create an attribute on the sub-device of type `SignalClass`,
        with a suffix of pv_suffix, which looks something like this::

            parent.sub.attribute_name = Cpt(SignalClass, pv_suffix, **keyword_arg_dict)

        Keep in mind that this is actually done in the metaclass creation, and
        not exactly as written above.
    clsname : str, optional
        The name of the class to be generated
        This defaults to {parent_name}{this_attribute_name.capitalize()}
    doc : str, optional
        The docstring to put on the dynamically generated class
    default_read_attrs : list, optional
        A class attribute to put on the dynamically generated class
    default_configuration_attrs : list, optional
        A class attribute to put on the dynamically generated class
    component_class : class, optional
        Defaults to Component
    base_class : class, optional
        Defaults to Device
    """

    def __init__(
        self,
        defn,
        *,
        clsname=None,
        doc=None,
        kind=Kind.normal,
        default_read_attrs=None,
        default_configuration_attrs=None,
        component_class=Component,
        base_class=None,
    ):
        if isinstance(default_read_attrs, Iterable):
            default_read_attrs = tuple(default_read_attrs)

        if isinstance(default_configuration_attrs, Iterable):
            default_configuration_attrs = tuple(default_configuration_attrs)

        self.defn = defn
        self.clsname = clsname
        self.default_read_attrs = default_read_attrs
        self.default_configuration_attrs = default_configuration_attrs
        self.attrs = list(defn.keys())
        self.component_class = component_class
        self.base_class = base_class if base_class is not None else Device
        self.components = {
            attr: component_class(cls, suffix, **kwargs)
            for attr, (cls, suffix, kwargs) in self.defn.items()
        }

        # NOTE: cls is None here, as it gets created in __set_name__, below
        super().__init__(cls=None, suffix="", lazy=False, kind=kind)

        # Allow easy access to all generated components directly in the
        # DynamicDeviceComponent instance
        for attr, cpt in self.components.items():
            if not hasattr(self, attr):
                setattr(self, attr, cpt)

    def __getnewargs_ex__(self):
        "Get arguments needed to copy this class (used for pickle/copy)"
        kwargs = dict(
            clsname=self.clsname,
            doc=self.doc,
            kind=self.kind,
            default_read_attrs=self.default_read_attrs,
            default_configuration_attrs=self.default_configuration_attrs,
            component_class=self.component_class,
            base_class=self.base_class,
        )
        return ((self.defn,), kwargs)

    def __set_name__(self, owner, attr_name):
        if self.clsname is None:
            self.clsname = underscores_to_camel_case(attr_name)
        super().__set_name__(owner, attr_name)
        self.cls = create_device_from_components(
            self.clsname,
            default_read_attrs=self.default_read_attrs,
            default_configuration_attrs=self.default_configuration_attrs,
            base_class=self.base_class,
            **self.components,
        )

    def __repr__(self):
        return "\n".join(f"{attr} = {cpt!r}" for attr, cpt in self.components.items())

    def subscriptions(self, event_type):
        raise NotImplementedError(
            "DynamicDeviceComponent does not yet " "support decorator subscriptions"
        )


# These stub 'Interface' classes are the apex of the mro heirarchy for
# their respective methods. They make multiple interitance more
# forgiving, and let us define classes that customize these methods
# but are not full Devices.


class BlueskyInterface:
    """Classes that inherit from this can safely customize the
    these methods without breaking mro.

    """

    def __init__(self, *args, **kwargs):
        # Subclasses can populate this with (signal, value) pairs, to be
        # set by stage() and restored back by unstage().
        self.stage_sigs = OrderedDict()

        self._staged = Staged.no
        self._original_vals = OrderedDict()
        super().__init__(*args, **kwargs)

    def trigger(self) -> StatusBase:
        """Trigger the device and return status object.

        This method is responsible for implementing 'trigger' or
        'acquire' functionality of this device.

        If there is an appreciable time between triggering the device
        and it being able to be read (via the
        :meth:`~BlueskyInterface.read` method) then this method is
        also responsible for arranging that the
        :obj:`~ophyd.status.StatusBase` object returned by this method
        is notified when the device is ready to be read.

        If there is no delay between triggering and being readable,
        then this method must return a :obj:`~ophyd.status.StatusBase`
        object which is already completed.

        Returns
        -------
        status : StatusBase
            :obj:`~ophyd.status.StatusBase` object which will be marked
            as complete when the device is ready to be read.

        """
        pass

    def read(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Read data from the device.

        This method is expected to be as instantaneous as possible,
        with any substantial acquisition time taken care of in
        :meth:`~BlueskyInterface.trigger`.

        The `OrderedDict` returned by this method must have identical
        keys (in the same order) as the `OrderedDict` returned by
        :meth:`~BlueskyInterface.describe()`.

        By convention, the first key in the return is the 'primary' key
        and maybe used by heuristics in :mod:`bluesky`.

        The values in the ordered dictionary must be dict (-likes) with the
        keys ``{'value', 'timestamp'}``.  The ``'value'`` may have any type,
        the timestamp must be a float UNIX epoch timestamp in UTC.

        Returns
        -------
        data : OrderedDict
            The keys must be strings and the values must be dict-like
            with the keys ``{'value', 'timestamp'}``

        """
        return OrderedDict()

    def describe(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Provide schema and meta-data for :meth:`~BlueskyInterface.read`.

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        return OrderedDict()

    def stage(self) -> List[object]:
        """Stage the device for data collection.

        This method is expected to put the device into a state where
        repeated calls to :meth:`~BlueskyInterface.trigger` and
        :meth:`~BlueskyInterface.read` will 'do the right thing'.

        Staging not idempotent and should raise
        :obj:`RedundantStaging` if staged twice without an
        intermediate :meth:`~BlueskyInterface.unstage`.

        This method should be as fast as is feasible as it does not return
        a status object.

        The return value of this is a list of all of the (sub) devices
        stage, including it's self.  This is used to ensure devices
        are not staged twice by the :obj:`~bluesky.run_engine.RunEngine`.

        This is an optional method, if the device does not need
        staging behavior it should not implement `stage` (or
        `unstage`).

        Returns
        -------
        devices : list
            list including self and all child devices staged

        """
        if self._staged == Staged.no:
            pass  # to short-circuit checking individual cases
        elif self._staged == Staged.yes:
            raise RedundantStaging(
                "Device {!r} is already staged. " "Unstage it first.".format(self)
            )
        elif self._staged == Staged.partially:
            raise RedundantStaging(
                "Device {!r} has been partially staged. "
                "Maybe the most recent unstaging "
                "encountered an error before finishing. "
                "Try unstaging again.".format(self)
            )
        self.log.debug("Staging %s", self.name)
        self._staged = Staged.partially

        # Resolve any stage_sigs keys given as strings: 'a.b' -> self.a.b
        stage_sigs = OrderedDict()
        for k, v in self.stage_sigs.items():
            if isinstance(k, str):
                # Device.__getattr__ handles nested attr lookup
                stage_sigs[getattr(self, k)] = v
            else:
                stage_sigs[k] = v

        # Read current values, to be restored by unstage()
        original_vals = {sig: sig.get() for sig in stage_sigs}

        # We will add signals and values from original_vals to
        # self._original_vals one at a time so that
        # we can undo our partial work in the event of an error.

        # Apply settings.
        devices_staged = []
        try:
            for sig, val in stage_sigs.items():
                self.log.debug(
                    "Setting %s to %r (original value: %r)",
                    sig.name,
                    val,
                    original_vals[sig],
                )
                sig.set(val).wait()
                # It worked -- now add it to this list of sigs to unstage.
                self._original_vals[sig] = original_vals[sig]
            devices_staged.append(self)

            # Call stage() on child devices.
            for attr in self._sub_devices:
                device = getattr(self, attr)
                if hasattr(device, "stage"):
                    device.stage()
                    devices_staged.append(device)
        except Exception:
            self.log.debug(
                "An exception was raised while staging %s or "
                "one of its children. Attempting to restore "
                "original settings before re-raising the "
                "exception.",
                self.name,
            )
            self.unstage()
            raise
        else:
            self._staged = Staged.yes
        return devices_staged

    def unstage(self) -> List[object]:
        """Unstage the device.

        This method returns the device to the state it was prior to the
        last `stage` call.

        This method should be as fast as feasible as it does not
        return a status object.

        This method must be idempotent, multiple calls (without a new
        call to 'stage') have no effect.

        Returns
        -------
        devices : list
            list including self and all child devices unstaged

        """
        self.log.debug("Unstaging %s", self.name)
        self._staged = Staged.partially
        devices_unstaged = []

        # Call unstage() on child devices.
        for attr in self._sub_devices[::-1]:
            device = getattr(self, attr)
            if hasattr(device, "unstage"):
                device.unstage()
                devices_unstaged.append(device)

        # Restore original values.
        for sig, val in reversed(list(self._original_vals.items())):
            self.log.debug("Setting %s back to its original value: %r", sig.name, val)
            sig.set(val).wait()
            self._original_vals.pop(sig)
        devices_unstaged.append(self)

        self._staged = Staged.no
        return devices_unstaged

    def pause(self) -> None:
        """Attempt to 'pause' the device.

        This is called when ever the
        :obj:`~bluesky.run_engine.RunEngine` is interrupted.

        A device may have internal state that means plans can not
        safely be re-wound.  This method may: put the device in a
        'paused' state and/or raise
        :obj:`~bluesky.run_engine.NoReplayAllowed` to indicate that
        the plan can not be rewound.

        Raises
        ------
        bluesky.run_engine.NoReplayAllowed

        """
        pass

    def resume(self) -> None:
        """Resume a device from a 'paused' state.

        This is called by the :obj:`bluesky.run_engine.RunEngine`
        when it resumes from an interruption and is responsible for
        ensuring that the device is ready to take data again.
        """
        pass

    def _validate_kind(self, val):
        return super()._validate_kind(val)


class GenerateDatumInterface:
    """Classes that inherit from this can safely customize the
    `generate_datum` method without breaking mro. If used along with the
    BlueskyInterface, inherit from this second."""

    def generate_datum(self, key, timestamp, datum_kwargs):
        pass


class Device(BlueskyInterface, OphydObject):
    """Base class for device objects

    This class provides attribute access to one or more Signals, which can be
    a mixture of read-only and writable. All must share the same base_name.

    Parameters
    ----------
    prefix : str, optional
        The PV prefix for all components of the device
    name : str, keyword only
        The name of the device (as will be reported via read()`
    kind : a member of the :class:`~ophydobj.Kind` :class:`~enum.IntEnum`
        (or equivalent integer), optional
        Default is ``Kind.normal``. See :class:`~ophydobj.Kind` for options.
    read_attrs : sequence of attribute names
        DEPRECATED: the components to include in a normal reading
        (i.e., in ``read()``)
    configuration_attrs : sequence of attribute names
        DEPRECATED: the components to be read less often (i.e., in
        ``read_configuration()``) and to adjust via ``configure()``
    parent : instance or None, optional
        The instance of the parent device, if applicable

    Attributes
    ----------
    lazy_wait_for_connection : bool
        When instantiating a lazy signal upon first access, wait for it to
        connect before returning control to the user.  See also the context
        manager helpers: ``wait_for_lazy_connection`` and
        ``do_not_wait_for_lazy_connection``.

    Subscriptions
    -------------
    SUB_ACQ_DONE
        A one-time subscription indicating the requested trigger-based
        acquisition has completed.
    """

    SUB_ACQ_DONE = "acq_done"  # requested acquire

    # Over-ride in sub-classes to control the default contents of read and
    # configuration attrs lists.
    # For read attributes, if `None`, defaults to `self.component_names'
    _default_read_attrs = None

    # For configuration attributes, if `None`, defaults to `[]`
    _default_configuration_attrs = None

    # When instantiating a lazy signal upon first access, wait for it to
    # connect before returning control to the user
    lazy_wait_for_connection = True

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        self._destroyed = False

        # Store EpicsSignal objects (only created once they are accessed)
        self._signals = {}

        # Copy the Device-defined signal kinds, for user modification
        self._component_kinds = self._component_kinds.copy()

        # Subscriptions to run or general methods necessary to call prior to
        # marking the Device as connected
        self._required_for_connection = self._required_for_connection.copy()

        self.prefix = prefix
        if self.component_names and prefix is None:
            raise ValueError("Must specify prefix if device signals are being " "used")

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)

        # The logic of these if blocks is:
        # - If the user did not provide read_attrs, fall back on the default
        # specified by the class, which ultimately falls back to Device, if no
        # subclass overrides it. Either way, we now have a read_attrs.
        # - If it is set to the sentinel ALL_COMPONENTS, ignore whatever the
        # 'kind' settings are; just include everything. This is an escape catch
        # for getting what _default_read_attrs=None used to do before 'kind'
        # was implemented.
        # - If it is set to a list, ignore whatever the 'kind' settings are and
        # just include that list.
        # - If it is set to None, respect whatever the 'kind' settings of the
        # components are.

        # If any sub-Devices are to be removed from configuration_attrs and
        # read_attrs, we have to remove them from read_attrs first, or they
        # will not allow themselves to be removed from configuration_attrs.
        if read_attrs is None:
            read_attrs = self._default_read_attrs
        if read_attrs is ALL_COMPONENTS:
            read_attrs = self.component_names
        if read_attrs is not None:
            self.read_attrs = list(read_attrs)

        if configuration_attrs is None:
            configuration_attrs = self._default_configuration_attrs
        if configuration_attrs is ALL_COMPONENTS:
            configuration_attrs = self.component_names
        if configuration_attrs is not None:
            self.configuration_attrs = list(configuration_attrs)

        with do_not_wait_for_lazy_connection(self):
            # Instantiate non-lazy signals and lazy signals with subscriptions
            [
                getattr(self, attr)
                for attr, cpt in self._sig_attrs.items()
                if not cpt.lazy or cpt._subscriptions
            ]

    @classmethod
    def _initialize_device(cls):
        """Initializes the Device and all of its Components

        Initializes the following attributes from the Components::
            - _sig_attrs - dict of attribute name to Component
            - component_names - a list of attribute names used for components
            - _device_tuple - An auto-generated namedtuple based on all
              existing Components in the Device
            - _sub_devices - a list of attributes which hold a Device
            - _required_for_connection - a dictionary of object-to-description
              for additional things that block this from being reported as
              connected
        """

        for attr in DEVICE_INSTANCE_ATTRS:
            if attr in cls.__dict__:
                raise TypeError(
                    "The attribute name %r is reserved for "
                    "use by the Device class. Choose a different "
                    "name." % attr
                )

        # this is so that the _sig_attrs class attribute includes the sigattrs
        # from all of its class-inheritance-parents so we do not have to do
        # this look up everytime we look at it.
        base_devices = [
            base for base in reversed(cls.__bases__) if hasattr(base, "_sig_attrs")
        ]

        cls._sig_attrs = OrderedDict(
            (attr, cpt)
            for base in base_devices
            for attr, cpt in base._sig_attrs.items()
            if getattr(cls, attr) is not None
        )

        # map component classes to their attribute names from this class
        this_sig_attrs = {
            attr: cpt
            for attr, cpt in cls.__dict__.items()
            if isinstance(cpt, Component)
        }

        cls._sig_attrs.update(**this_sig_attrs)

        # Record the class-defined kinds - these can be updated on a
        # per-instance basis
        cls._component_kinds = {attr: cpt.kind for attr, cpt in cls._sig_attrs.items()}

        bad_attrs = set(cls._sig_attrs).intersection(DEVICE_RESERVED_ATTRS)
        if bad_attrs:
            raise TypeError(
                f"The attribute name(s) {bad_attrs} are part of"
                f" the bluesky interface and cannot be used as "
                f"component names. Choose a different name."
            )

        # List Signal attribute names.
        cls.component_names = tuple(cls._sig_attrs)

        # The namedtuple associated with the device
        cls._device_tuple = namedtuple(
            f"{cls.__name__}Tuple",
            [comp for comp in cls.component_names[:254] if not comp.startswith("_")],
        )

        # List the attributes that are Devices (not Signals).
        # This list is used by stage/unstage. Only Devices need to be staged.
        cls._sub_devices = [
            attr for attr, cpt in cls._sig_attrs.items() if cpt.is_device
        ]

        # All (obj, description) that may block the Device from being shown as
        # connected:
        cls._required_for_connection = dict(
            obj._required_for_connection
            for attr, obj in cls.__dict__.items()
            if getattr(obj, "_required_for_connection", False)
        )

    def __init_subclass__(cls, **kwargs):
        "This is called automatically in Python for all subclasses of Device"
        super().__init_subclass__(**kwargs)
        cls._initialize_device()

    @classmethod
    def walk_components(cls):
        """Walk all components in the Device hierarchy

        Yields
        ------
        ComponentWalk
            Where ancestors is all ancestors of the signal, including the
            top-level device `walk_components` was called on.
        """
        for attr, cpt in cls._sig_attrs.items():
            yield ComponentWalk(ancestors=(cls,), dotted_name=attr, item=cpt)
            if issubclass(cpt.cls, Device) or hasattr(cpt.cls, "walk_components"):
                sub_dev = cpt.cls
                for walk in sub_dev.walk_components():
                    ancestors = (cls,) + walk.ancestors
                    dotted_name = ".".join((attr, walk.dotted_name))
                    yield ComponentWalk(
                        ancestors=ancestors, dotted_name=dotted_name, item=walk.item
                    )

    def walk_signals(self, *, include_lazy=False):
        """Walk all signals in the Device hierarchy

        EXPERIMENTAL: This method is experimental, and there are tentative
        plans to change its API in a way that may not be backward-compatible.

        Parameters
        ----------
        include_lazy : bool, optional
            Include not-yet-instantiated lazy signals

        Yields
        ------
        ComponentWalk
            Where ancestors is all ancestors of the signal, including the
            top-level device `walk_signals` was called on.
        """
        for attr, cpt in self._sig_attrs.items():
            # 2 scenarios:
            #  - Always include non-lazy components
            #  - Include a lazy if already instantiated OR requested with
            #    include_lazy
            lazy_ok = cpt.lazy and (include_lazy or attr in self._signals)
            should_walk = not cpt.lazy or lazy_ok

            if not should_walk:
                continue

            sig = getattr(self, attr)
            if isinstance(sig, Device):
                for walk in sig.walk_signals(include_lazy=include_lazy):
                    ancestors = (self,) + walk.ancestors
                    dotted_name = ".".join((attr, walk.dotted_name))
                    yield ComponentWalk(
                        ancestors=ancestors, dotted_name=dotted_name, item=walk.item
                    )
            else:
                yield ComponentWalk(ancestors=(self,), dotted_name=attr, item=sig)

    @classmethod
    def walk_subdevice_classes(cls):
        """Walk all sub-Devices classes in the Device hierarchy

        Yields
        ------
        (dotted_name, subdevice_class)
        """
        for attr in cls._sub_devices:
            cpt = getattr(cls, attr)
            if cpt is None:
                # Subclasses can override this, making this None...
                continue

            yield (attr, cpt.cls)
            for sub_attr, sub_cls in cpt.cls.walk_subdevice_classes():
                yield (".".join((attr, sub_attr)), sub_cls)

    def walk_subdevices(self, *, include_lazy=False):
        """Walk all sub-Devices in the hierarchy

        EXPERIMENTAL: This method is experimental, and there are tentative
        plans to change its API in a way that may not be backward-compatible.

        Yields
        ------
        (dotted_name, subdevice_instance)
        """
        # TODO: Devices can be lazy, outside of original design intent; should
        # discuss this at some point
        cls = type(self)
        for attr in cls._sub_devices:
            cpt = getattr(cls, attr)
            lazy_ok = cpt.lazy and (include_lazy or attr in self._signals)
            should_walk = not cpt.lazy or lazy_ok

            if should_walk:
                dev = getattr(self, attr)
                yield (attr, dev)
                for sub_attr, sub_dev in dev.walk_subdevices(include_lazy=include_lazy):
                    yield (".".join((attr, sub_attr)), sub_dev)

    def destroy(self):
        "Disconnect and destroy all signals on the Device"
        self._destroyed = True
        exceptions = []
        for walk in self.walk_signals(include_lazy=False):
            sig = walk.item
            try:
                sig.destroy()
            except Exception as ex:
                ex.signal = sig
                ex.attr = walk.dotted_name
                exceptions.append(ex)

        if exceptions:
            msg = ", ".join(
                "{} ({})".format(ex.attr, ex.__class__.__name__) for ex in exceptions
            )
            raise ExceptionBundle(
                "Failed to disconnect all signals ({})".format(msg),
                exceptions=exceptions,
            )
        super().destroy()

    def _get_kind(self, name):
        """Get a Kind for a given Component

        If the Component is instantiated, it will be retrieved directly from
        that object.

        If the Component is lazy and not yet instantiated, the default value as
        specified by the Component class will be used.  This is stashed away in
        `_component_kinds`.
        """
        return (
            self._signals[name].kind
            if name in self._signals
            else self._component_kinds[name]
        )

    def _set_kind(self, name, kind):
        """Set the Kind for a given Component"""
        if name in self._signals:
            self._signals[name].kind = kind
        else:
            self._component_kinds[name] = kind

    def _get_components_of_kind(self, kind):
        "Get names of components that match a specific kind"
        for component_name in self.component_names:
            # this might be lazy and get the Cpt
            component_kind = self._get_kind(component_name)
            if kind & component_kind:
                yield component_name, getattr(self, component_name)

    def _validate_kind(self, val):
        val = super()._validate_kind(val)
        if Kind.normal & val:
            val |= Kind.config
        return val

    @property
    def read_attrs(self):
        return self.OphydAttrList(self, Kind.normal, Kind.hinted, "read_attrs")

    @read_attrs.setter
    def read_attrs(self, val):
        self.__set_kinds_according_to_list(val, Kind.normal, Kind.hinted, "read_attrs")

    @property
    def configuration_attrs(self):
        return self.OphydAttrList(self, Kind.config, Kind.config, "configuration_attrs")

    @configuration_attrs.setter
    def configuration_attrs(self, val):
        self.__set_kinds_according_to_list(
            val, Kind.config, Kind.config, "configuration_attrs"
        )

    def __set_kinds_according_to_list(
        self, master_list, set_kind, unset_kind, recurse_name
    ):
        master_list = set(master_list)
        component_names = set(self.component_names)

        # For all children of this class, update their kinds
        for name in component_names:
            kind = self._get_kind(name)
            kind = kind | set_kind if name in master_list else kind & ~unset_kind
            self._set_kind(name, kind)

        # Now look at everything else, presumably things with dots
        extra = master_list - component_names
        fail = set(c for c in extra if "." not in c)

        if fail:
            raise ValueError(
                f"You asked to add the components {fail} to {recurse_name} "
                f"on {self.name!r}, but there is no such child in "
                f"{self.component_names}."
            )

        group = itertools.groupby(
            ((child, rest) for child, _, rest in (c.partition(".") for c in extra)),
            lambda x: x[0],
        )

        # we are into grand-children, can not be lazy!
        for child, cf_list in group:
            cpt = getattr(self, child)
            cpt.kind |= set_kind
            setattr(cpt, recurse_name, [c[1] for c in cf_list])

    @property
    def signal_names(self):
        warnings.warn(
            "'signal_names' has been renamed 'component_names' for "
            "clarity because it may include a mixture of Signals "
            "and Devices -- any Components. This alias may be "
            "removed in a future release of ophyd.",
            stacklevel=2,
        )
        return self.component_names

    def summary(self):
        print(self._summary())

    def _summary(self):
        "Return a string summarizing the structure of the Device."
        read_attrs = self.read_attrs
        config_attrs = self.configuration_attrs
        used_attrs = set(read_attrs + config_attrs)
        extra_attrs = [a for a in self.component_names if a not in used_attrs]
        hints = getattr(self, "hints", {}).get("fields", [])

        def format_leaf(attr):
            cpt = getattr(self, attr)
            return f"{attr:<20} {type(cpt).__name__:<20}({cpt.name!r})"

        def format_hint(key):
            return "".join(("*" if key in hints else " ", key))

        categories = [
            ("data keys (* hints)", sorted(self.describe()), format_hint),
            ("read attrs", read_attrs, format_leaf),
            ("config keys", sorted(self.describe_configuration()), str),
            ("configuration attrs", config_attrs, format_leaf),
            ("unused attrs", extra_attrs, format_leaf),
        ]

        out = []
        for title, items, formatter in categories:
            out.append(title)
            out.append("-" * len(title))
            out.extend([formatter(s) for s in items])
            out.append("")

        return "\n".join(out)

    def wait_for_connection(self, all_signals=False, timeout=2.0):
        """Wait for signals to connect

        Parameters
        ----------
        all_signals : bool, optional
            Wait for all signals to connect (including lazy ones)
        timeout : float or None
            Overall timeout
        """
        signals = [walk.item for walk in self.walk_signals(include_lazy=all_signals)]

        pending_funcs = {
            dev: getattr(dev, "_required_for_connection", {})
            for name, dev in self.walk_subdevices(include_lazy=all_signals)
        }
        pending_funcs[self] = self._required_for_connection

        t0 = ttime.time()
        while timeout is None or (ttime.time() - t0) < timeout:
            connected = all(sig.connected for sig in signals)
            if connected and not any(pending_funcs.values()):
                return
            ttime.sleep(min((0.05, timeout / 10.0)))

        def get_name(sig):
            sig_name = f"{self.name}.{sig.dotted_name}"
            return f"{sig_name} ({sig.pvname})" if hasattr(sig, "pvname") else sig_name

        reasons = []
        unconnected = ", ".join(get_name(sig) for sig in signals if not sig.connected)
        if unconnected:
            reasons.append(f"Failed to connect to all signals: {unconnected}")
        if any(pending_funcs.values()):
            pending = ", ".join(
                description.format(device=dev)
                for dev, funcs in pending_funcs.items()
                for obj, description in funcs.items()
            )
            reasons.append(f"Pending operations: {pending}")
        raise TimeoutError("; ".join(reasons))

    def get_instantiated_signals(self, *, attr_prefix=None):
        """Yields all of the instantiated signals in a device hierarchy

        Parameters
        ----------
        attr_prefix : string, optional
            The attribute prefix. If None, defaults to self.name

        Yields
        ------
            (fully_qualified_attribute_name, signal_instance)
        """
        if attr_prefix is None:
            attr_prefix = self.name

        for attr, sig in self._signals.items():
            # fully qualified attribute name from top-level device
            full_attr = "{}.{}".format(attr_prefix, attr)
            if isinstance(sig, Device):
                yield from sig.get_instantiated_signals(attr_prefix=full_attr)
            else:
                yield full_attr, sig

    @property
    def connected(self):
        signals_connected = all(
            walk.item.connected for walk in self.walk_signals(include_lazy=False)
        )
        pending_funcs = any(
            getattr(item, "_required_for_connection", None)
            for name, item in self.walk_subdevices()
        )

        pending_funcs = pending_funcs or self._required_for_connection
        return signals_connected and not pending_funcs

    def __getattr__(self, name):
        """Get a component from a fully-qualified name"""
        if "." in name:
            return operator.attrgetter(name)(self)

        # Components will be instantiated through the descriptor mechanism in
        # the Component class, so anything reaching this point is an error.
        raise AttributeError(name)

    def _instantiate_component(self, attr):
        "Create a Component specifically for this Device"
        if self._destroyed:
            raise RuntimeError("Cannot instantiate new signals on a destroyed Device")

        try:
            # Initial access of signal
            cpt = self._sig_attrs[attr]
        except KeyError:
            raise RuntimeError(
                f"The Component {attr!r} exists at the Python level and "
                "has triggered the `_instantiate_component` "
                "code path on Device, but has not been registered with "
                "the Component management machinery in Device.  This may be due to "
                "using multiple inheritance with a mix-in class that defines "
                "a Component but does not inherent from Device."
            ) from None

        try:
            self._signals[attr] = cpt.create_component(self)
            sig = self._signals[attr]
            for event_type, functions in cpt._subscriptions.items():
                for func in functions:
                    method = getattr(self, func.__name__)
                    sig.subscribe(method, event_type=event_type, run=sig.connected)
        except AttributeError as ex:
            # Raise a different Exception, as AttributeError will be shadowed
            # during initial access
            raise RuntimeError(
                f"AttributeError while instantiating " f"component: {attr}"
            ) from ex

        return sig

    @doc_annotation_forwarder(BlueskyInterface)
    def read(self):
        res = super().read()

        for _, component in self._get_components_of_kind(Kind.normal):
            res.update(component.read())
        return res

    def read_configuration(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Dictionary mapping names to value dicts with keys: value, timestamp

        To control which fields are included, change the Component kinds on the
        device, or modify the ``configuration_attrs`` list.
        """
        res = OrderedDict()

        for _, component in self._get_components_of_kind(Kind.config):
            res.update(component.read_configuration())
        return res

    @doc_annotation_forwarder(BlueskyInterface)
    def describe(self):
        res = super().describe()
        for _, component in self._get_components_of_kind(Kind.normal):
            res.update(component.describe())
        return res

    def describe_configuration(self) -> OrderedDictType[str, Dict[str, Any]]:
        """Provide schema & meta-data for :meth:`~BlueskyInterface.read_configuration`

        This keys in the `OrderedDict` this method returns must match the
        keys in the `OrderedDict` return by :meth:`~BlueskyInterface.read`.

        This provides schema related information, (ex shape, dtype), the
        source (ex PV name), and if available, units, limits, precision etc.

        Returns
        -------
        data_keys : OrderedDict
            The keys must be strings and the values must be dict-like
            with the ``event_model.event_descriptor.data_key`` schema.
        """
        res = OrderedDict()
        for _, component in self._get_components_of_kind(Kind.config):
            res.update(component.describe_configuration())
        return res

    @property
    def hints(self):
        fields = []
        for _, component in self._get_components_of_kind(Kind.normal):
            c_hints = component.hints
            fields.extend(c_hints.get("fields", []))
        return {"fields": fields}

    @property
    def trigger_signals(self):
        names = [
            attr
            for attr, cpt in self._sig_attrs.items()
            if cpt.trigger_value is not None
        ]

        return [getattr(self, name) for name in names]

    def _done_acquiring(self, **kwargs):
        """Call when acquisition has completed."""
        self._run_subs(sub_type=self.SUB_ACQ_DONE, success=True, **kwargs)
        self._reset_sub(self.SUB_ACQ_DONE)

    @doc_annotation_forwarder(BlueskyInterface)
    def trigger(self):
        """Start acquisition"""
        signals = self.trigger_signals
        if len(signals) > 1:
            raise NotImplementedError(
                "More than one trigger signal is not " "currently supported"
            )
        status = DeviceStatus(self)
        if not signals:
            status.set_finished()
            return status

        (acq_signal,) = signals

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def done_acquisition(**ignored_kwargs):
            # Keyword arguments are ignored here from the EpicsSignal
            # subscription, as the important part is that the put completion
            # has finished
            self._done_acquiring()

        acq_signal.put(1, wait=False, callback=done_acquisition)
        return status

    def stop(self, *, success=False):
        """Stop the Device and all (instantiated) subdevices"""
        exc_list = []

        for attr, dev in getattrs(self, self._sub_devices):
            if not dev.connected:
                self.log.debug(
                    "stop: device %s (%s) is not connected; " "skipping", attr, dev
                )
                continue

            try:
                dev.stop(success=success)
            except ExceptionBundle as ex:
                exc_list.extend(
                    [
                        ("{}.{}".format(attr, sub_attr), ex)
                        for sub_attr, ex in ex.exceptions.items()
                    ]
                )
            except Exception as ex:
                exc_list.append((attr, ex))
                self.log.exception("Device %s (%s) stop failed", attr, dev)

        if exc_list:
            exc_info = "\n".join(
                "{} raised {!r}".format(attr, ex) for attr, ex in exc_list
            )
            raise ExceptionBundle(
                "{} exception(s) were raised during stop: \n"
                "{}".format(len(exc_list), exc_info),
                exceptions=dict(exc_list),
            )

    def get(self, **kwargs):
        """Get the value of all components in the device

        Keyword arguments are passed onto each signal.get(). Components
        beginning with an underscore will not be included.
        """
        values = {}
        for attr in self.component_names:
            if not attr.startswith("_"):
                signal = getattr(self, attr)
                values[attr] = signal.get(**kwargs)

        return self._device_tuple(**values)

    def put(self, dev_t, **kwargs):
        """Put a value to all components of the device

        Keyword arguments are passed onto each signal.put()

        Parameters
        ----------
        dev_t : DeviceTuple or tuple
            The device tuple with the value(s) to put (see get_device_tuple)
        """
        if not isinstance(dev_t, self._device_tuple):
            try:
                dev_t = self._device_tuple(*dev_t)
            except TypeError as ex:
                raise ValueError(
                    "{}\n\tDevice tuple fields: {}"
                    "".format(ex, self._device_tuple._fields)
                )

        for attr in self.component_names:
            value = getattr(dev_t, attr)
            signal = getattr(self, attr)
            signal.put(value, **kwargs)

    @classmethod
    def get_device_tuple(cls):
        """The device tuple type associated with an Device class

        This is a tuple representing the full state of all components and
        dynamic device sub-components.
        """
        return cls._device_tuple

    def configure(self, d: Dict[str, Any]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Configure the device for something during a run

        This default implementation allows the user to change any of the
        `configuration_attrs`. Subclasses might override this to perform
        additional input validation, cleanup, etc.

        Parameters
        ----------
        d : dict
            The configuration dictionary. To specify the order that
            the changes should be made, use an OrderedDict.

        Returns
        -------
        (old, new) tuple of dictionaries
        Where old and new are pre- and post-configure configuration states.
        """
        old = self.read_configuration()
        for key, val in d.items():
            if key not in self.configuration_attrs:
                # a little extra checking for a more specific error msg
                if key not in self.component_names:
                    raise ValueError("There is no signal named %s" % key)
                else:
                    raise ValueError(
                        "%s is not one of the "
                        "configuration_fields, so it cannot be "
                        "changed using configure" % key
                    )
            getattr(self, key).set(val).wait()
        new = self.read_configuration()
        return old, new

    def _repr_info(self):
        yield ("prefix", self.prefix)
        yield from super()._repr_info()

        yield ("read_attrs", self.read_attrs)
        yield ("configuration_attrs", self.configuration_attrs)

    class OphydAttrList(MutableSequence):
        """list proxy to migrate away from Device.read_attrs and Device.config_attrs"""

        def __init__(self, device, kind, remove_kind, recurse_key):
            self._kind = kind
            self._remove_kind = remove_kind
            self._parent = device
            self._recurse_key = recurse_key

        def __internal_list(self):
            def per_component(name, component):
                yield name
                for v in getattr(component, self._recurse_key, []):
                    yield ".".join([name, v])

            out = (
                per_component(name, component)
                for name, component in self._parent._get_components_of_kind(self._kind)
            )

            return list(itertools.chain.from_iterable(out))

        def __getitem__(self, key):
            return self.__internal_list()[key]

        def __setitem__(self, key, val):
            raise NotImplementedError

        def __delitem__(self, key):
            to_delete = self.__internal_list()[key]
            if not isinstance(key, slice):
                to_delete = [to_delete]

            for attr in to_delete:
                item_kind = self._parent._get_kind(attr)
                item_kind &= ~self._remove_kind

        def __len__(self):
            return len(self.__internal_list())

        def insert(self, index, object):
            getattr(self._parent, object).kind |= self._kind

        def remove(self, value):
            kind = self._parent._get_kind(value)
            kind &= ~self._remove_kind

        def __contains__(self, value):
            return value in self.__internal_list()

        def __iter__(self):
            yield from self.__internal_list()

        def __eq__(self, other):
            return list(self) == other

        def __repr__(self):
            return repr(list(self))

        def __add__(self, other):
            return list(self) + list(other)

        def __radd__(self, other):
            return list(self) + list(other)


# Device can be used on its own in trivial cases; ensure that it is ready
# out-of-the-box for this scenario.
if not hasattr(Device, "_sig_attrs"):
    Device._initialize_device()


@contextlib.contextmanager
def kind_context(kind):
    yield functools.partial(Component, kind=kind)


def create_device_from_components(
    name,
    *,
    docstring=None,
    default_read_attrs=None,
    default_configuration_attrs=None,
    base_class=Device,
    class_kwargs=None,
    **components,
):
    """Factory function to make a Device from Components

    Parameters
    ----------
    name : str
        Class name to create
    docstring : str, optional
        Docstring to attach to the class
    default_read_attrs : list, optional
        Outside of Kind, control the default read_attrs list.
        Defaults to all `component_names'
    default_configuration_attrs : list, optional
        Outside of Kind, control the default configuration_attrs list.
        Defaults to []
    base_class : Device or sub-class, optional
        Class to inherit from, defaults to Device
    **components : dict
        Keyword arguments are used to map component attribute names to
        Components.

    Returns
    -------
    cls : Device
        Newly generated Device class
    """
    if docstring is None:
        docstring = f"{name} Device"

    if not isinstance(base_class, tuple):
        base_class = (base_class,)

    if class_kwargs is None:
        class_kwargs = {}

    clsdict = OrderedDict(
        __doc__=docstring,
        _default_read_attrs=default_read_attrs,
        _default_configuration_attrs=default_configuration_attrs,
    )

    for attr, component in components.items():
        if not isinstance(component, Component):
            raise ValueError(
                f"Attribute {attr} is not a Component. "
                f"It is of type {type(component).__name__}"
            )

        clsdict[attr] = component

    return type(name, base_class, clsdict, **class_kwargs)


def required_for_connection(func=None, *, description=None, device=None):
    """Require that a method be called prior to marking a Device as connected

    This is a decorator that wraps the given function.  When the function is
    called for the first time, the Device instance is informed that this
    operation is no longer pending.

    Parameters
    ----------
    func : callable, optional
        Function to wrap
    description : str, optional
        Optional string description to be shown when `wait_for_connection`
        fails.  Can include {device} which will be substituted when presented
        to the user.
    """

    if func is None:
        if description is None:
            raise ValueError("Either func or description must be specified")
        return functools.partial(required_for_connection, description=description)

    if description is None:
        if hasattr(func, "_subscriptions"):
            description = ", ".join(
                f"{{device}}.{func.__name__}[{event_type}] subscription"
                for cpt, event_type in func._subscriptions
            )
        else:
            description = f"{func.__name__} call"

    key = inspect.unwrap(func)

    if device is not None:
        # With the Device specified, this can only be done post-init, with the
        # required_for_connection flag set prior to returning.
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            try:
                ret = func(*args, **kwargs)
            finally:
                device._required_for_connection.pop(key, None)
            return ret

        # Add a specific requirement
        device._required_for_connection[func] = description
    else:
        # With the Device unspecified, this can only be used as a decorator on
        # unbound methods.
        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            try:
                ret = func(self, *args, **kwargs)
            finally:
                self._required_for_connection.pop(key, None)
            return ret

    wrapped._required_for_connection = (key, description)
    return wrapped


def _wait_for_connection_context(value, doc):
    @contextlib.contextmanager
    def wrapped(dev):

        orig = dev.lazy_wait_for_connection
        dev.lazy_wait_for_connection = value
        try:
            yield
        finally:
            dev.lazy_wait_for_connection = orig

    wrapped.__doc__ = f"""Context manager which changes the wait behavior of lazy signal instantiation

By default, upon instantiation of a lazy signal, `wait_for_connection`
is called.  While a common source of confusion, this is done
intentionally and for good reason: without this functionality in place,
any new lazy signal will generally take a finite amount of time to
connect. This then requires that the user manually call
`wait_for_connection` each time before using the signal.

In certain cases, it can be desirable to override this behavior. For
instance, when instantiating multiple lazy signals or instantiating a
signal just so that a subscription can be added.

{doc}

Parameters
----------
dev : Device
    The device to temporarily change
"""
    return wrapped


wait_for_lazy_connection = _wait_for_connection_context(
    True, doc="Wait for lazy signals to connect post-instantiation."
)
do_not_wait_for_lazy_connection = _wait_for_connection_context(
    False, doc="Do not wait for lazy signals to connect post-instantiation."
)
